from ._version import version_info, __version__

from .pysbviz import pysbViz
from .views import *


def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': 'static',
        'dest': 'pyvipr',
        'require': 'pyvipr/extension'
    }]


__all__ = ['pysbViz'] + views.__all__



from collections import OrderedDict
import networkx as nx
import json
from .util_networkx import from_networkx
import collections
import pysb
import pyvipr.util as hf
from pysb.bng import generate_equations
from networkx.algorithms import bipartite
from community import best_partition


class OrderedGraph(nx.DiGraph):
    """
    Networkx Digraph that stores the nodes in the order they are input
    """
    node_dict_factory = OrderedDict
    adjlist_outer_dict_factory = OrderedDict
    adjlist_inner_dict_factory = OrderedDict
    edge_attr_dict_factory = OrderedDict


class StaticViz(object):
    """
    Class to visualize models and their dynamics

    Parameters
    ----------
    model : pysb.Model
        PySB Model to visualize.
    """

    def __init__(self, model):
        # Need to create a model visualization base and then do independent visualizations: static and dynamic
        self.model = model
        self.graph = None
        generate_equations(self.model)

    # This is a function to merge several nodes into one in a Networkx graph
    @staticmethod
    def merge_nodes(G, nodes, new_node, **attr):
        """
        Merges the selected `nodes` of the graph G into one `new_node`,
        meaning that all the edges that pointed to or from one of these
        `nodes` will point to or from the `new_node`.
        attr_dict and **attr are defined as in `G.add_node`.
        """
        G.add_node(new_node, **attr)  # Add the 'merged' node
        newG = G.copy()

        for n1, n2, data in newG.edges(data=True):
            # For all edges related to one of the nodes to merge,
            # make an edge going to or coming from the `new gene`.
            if n1 in nodes:
                G.add_edge(new_node, n2, **data)
            elif n2 in nodes:
                G.add_edge(n1, new_node, **data)

        for n in nodes:  # remove the merged nodes
            G.remove_node(n)

    def species_view(self):
        """
        Generate data to create a species network

        Parameters
        ----------

        Examples
        --------
        >>> from pysb.examples.earm_1_0 import model
        >>> viz = StaticViz(model)
        >>> data = viz.species_view()

        Returns
        -------
        A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
        contains all the information (nodes,edges, positions) to generate a cytoscapejs network.
        """
        graph = self.species_graph()
        g_layout = dot_layout(graph)
        data = graph_to_json(sp_graph=graph, layout=g_layout)
        return data

    def species_compartments_view(self):
        """
        Generate data to create a species networks and their respective compartments
        Returns
        -------
        A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
        contains all the information (nodes, parent_nodes, edges, positions) to
         generate a cytoscapejs network.
        """
        graph = self.compartments_data_graph()
        g_layout = dot_layout(graph)
        data = graph_to_json(sp_graph=graph, layout=g_layout)
        return data

    def compartments_data_graph(self):
        # Check if model has comparments
        if not self.model.compartments:
            raise ValueError('Model has no compartments')
        graph = self.species_graph()
        compartment_nodes = []
        for c in self.model.compartments:
            if c.parent is not None:
                compartment_nodes.append((c.name, dict(parent=c.parent.name)))
            else:
                compartment_nodes.append(c.name)

        graph.add_nodes_from(compartment_nodes, NodeType='compartment')
        # Determining compartment node family tree
        # Determine species compartment
        sp_compartment = {}
        for idx, sp in enumerate(self.model.species):
            monomers = sp.monomer_patterns
            monomers_comp = {m.compartment.name: m.compartment.size for m in monomers}
            smallest_comp = min(monomers_comp, key=monomers_comp.get)
            sp_compartment['s{0}'.format(idx)] = smallest_comp
        nx.set_node_attributes(graph, sp_compartment, 'parent')
        return graph

    def communities_view(self, random_state=None):
        """
        Use a community detection algorithm to find groups of nodes that are densely connected.
        It generates the data to create a network with compound nodes that hold the communities.
        Returns
        -------
        A Dictionary object that can be converted into Cytoscape.js JSON. This dictionary
        contains all the information (nodes,edges, parent nodes, positions) to generate
        a cytoscapejs network.
        """
        graph = self.communities_data_graph(random_state=random_state)
        g_layout = dot_layout(graph)
        data = graph_to_json(sp_graph=graph, layout=g_layout)
        return data

    def communities_data_graph(self, random_state=None):
        graph = self.species_graph()
        graph_communities = graph.copy().to_undirected()  # Louvain algorithm only deals with undirected graphs
        communities = best_partition(graph_communities, random_state=random_state)
        # compound nodes to add to hold communities
        cnodes = set(communities.values())
        graph.add_nodes_from(cnodes, NodeType='community')
        nx.set_node_attributes(graph, communities, 'parent')
        return graph

    def sp_rxns_bidirectional_view(self):
        """
        Generate a dictionary with the info of a bipartite graph where one set of
        nodes is the model species and the other set is the model bidirectional reactions

        Parameters
        ----------

        Returns
        -------
        A dictionary object with the graph information
        """
        graph = self.sp_rxns_bidirectional_graph()
        g_layout = dot_layout(graph)
        data = graph_to_json(sp_graph=graph, layout=g_layout)
        return data

    def sp_rxns_view(self):
        """
        Generates a dictionary with the info of a bipartite graph where one set of nodes is the model species
        and the other set is the unidirectional reactions
        Returns
        -------
        A dictionary object with the graph information.
        """
        graph = self.sp_rxns_graph()
        g_layout = dot_layout(graph)
        data = graph_to_json(sp_graph=graph, layout=g_layout)
        return data

    def sp_rules_view(self):
        """
        Generates a dictionary with the info of a bipartite graph where one set of nodes is the model species
        and the other set is the model rules
        Returns
        -------

        """
        rules_graph = self.rules_graph()
        rules_graph = self.graph_merged_pair_edges(rules_graph)
        g_layout = dot_layout(rules_graph)
        data = graph_to_json(sp_graph=rules_graph, layout=g_layout)
        return data

    def sp_rules_functions_view(self):
        """
        Generates a dictionary with the info of a bipartite graph where one set of nodes is the model species
        and the other set is the model rules. Additionally, it adds information of the functions from which
        the rules come from.
        Returns
        -------

        """
        rules_graph = self.rules_graph()
        rules_graph = self.graph_merged_pair_edges(rules_graph)
        rule_functions = {rule.name: rule._function for rule in self.model.rules}
        unique_functions = set(rule_functions.values())
        nx.set_node_attributes(rules_graph, rule_functions, 'parent')
        rules_graph.add_nodes_from(unique_functions, NodeType='function')
        g_layout = dot_layout(rules_graph)
        data = graph_to_json(sp_graph=rules_graph, layout=g_layout)
        return data

    def sp_rules_modules_view(self):
        """
        Generates a dictionary with the info of a bipartite graph where one set of nodes is the model species
        and the other set is the model rules. Additionally, it adds information of the modules from which
        the rules come from.
        Returns
        -------

        """
        rules_graph = self.rules_graph()
        rules_graph = self.graph_merged_pair_edges(rules_graph)

        # Unique modules used in the model
        unique_modules = [m for m in self.model.modules if not m.startswith('_')]
        # Remove outer model file to not look further modules that are not related to
        # the model
        unique_modules.remove(self.model.name)
        module_parents = {}
        # pysb components have a modules list that starts from the inner frame
        # to the outer frame.
        for module in unique_modules:
            total_parents = []
            for rule in self.model.rules:
                mods = rule._modules
                try:
                    mod_idx = mods.index(module)
                    module_parent = mods[mod_idx + 1]
                except (ValueError, IndexError):
                    continue
                total_parents.append(module_parent)

            total_parents = list(set(total_parents))
            if module in total_parents:
                total_parents.remove(module)

            if len(total_parents) >= 2:
                for idx, p in enumerate(total_parents):
                    module_parents[module + str(idx)] = p
            else:
                module_parents[module] = total_parents[0]

        rules_module = {}
        for rule in self.model.rules:
            m = rule._modules[0]
            if m not in module_parents.keys():
                m_parent = rule._modules[1]
                m_parent_child = list(module_parents.keys())[list(module_parents.values()).index(m_parent)]
                rules_module[rule.name] = m_parent_child
            else:
                rules_module[rule.name] = m

        module_parent_nodes = list(module_parents.keys())
        module_parent_nodes.append(self.model.name)
        rules_graph.add_nodes_from(module_parent_nodes, NodeType='module')
        nx.set_node_attributes(rules_graph, module_parents, 'parent')
        nx.set_node_attributes(rules_graph, rules_module, 'parent')
        g_layout = dot_layout(rules_graph)
        data = graph_to_json(sp_graph=rules_graph, layout=g_layout)
        return data

    def projections_view(self, project_to='species_reactions'):
        """
        Generates a dictionary with the info of a graph representing the PySB model.

        Parameters
        ----------
        project_to: str
            Nodes to which the graph is going to be projected. Options are:
            'species_reactions', 'reactions', 'rules', 'species_rules'

        Returns
        -------
        A dictionary object with the graph information

        """
        if project_to == 'species_reactions' or project_to == 'reactions':
            bipartite_graph = self.sp_rxns_bidirectional_graph()
        elif project_to == 'rules' or project_to == 'species_rules':
            bipartite_graph = self.rules_graph()
        else:
            raise ValueError('Projection not valid')

        projected_graph = self.projected_graph(bipartite_graph, project_to)
        g_layout = dot_layout(projected_graph)
        data = graph_to_json(sp_graph=projected_graph, layout=g_layout)
        return data

    def projected_species_reactions_view(self):
        return self.projections_view('species_reactions')

    def projected_reactions_view(self):
        return self.projections_view('reactions')

    def projected_rules_view(self):
        return self.projections_view('rules')

    def projected_species_rules_view(self):
        return self.projections_view('species_rules')

    def _sp_initial(self, sp):
        """
        Get initial condition of a species
        Parameters
        ----------
        sp: pysb.ComplexPattern, pysb species

        Returns
        -------

        """
        sp_0 = 0
        for spInitial in self.model.initial_conditions:
            if spInitial[0].is_equivalent_to(sp):
                sp_0 = spInitial[1].value
                break
        return sp_0

    def species_graph(self):
        """
        Creates a nx.DiGraph graph of the model species

        Returns
        -------
        A :class: nx.Digraph graph that has the information for the visualization of the model
        """
        sp_graph = OrderedGraph(name=self.model.name, graph={'rankdir':'LR'}, paths=[])

        # TODO: there are reactions that generate parallel edges that are not taken into account because netowrkx
        # digraph only allows one edge between two nodes
        ic_species = [cp for cp, parameter in self.model.initial_conditions]
        for idx, sp in enumerate(self.model.species):
            species_node = 's%d' % idx
            color = "#2b913a"
            # color species with an initial condition differently
            if len([s for s in ic_species if s.is_equivalent_to(sp)]):
                color = "#aaffff"
            # Setting the information about the node
            node_data = dict(label=hf.parse_name(sp),
                             background_color=color,
                             shape='ellipse',
                             NodeType='species',
                             spInitial=self._sp_initial(sp))
            sp_graph.add_node(species_node, **node_data)

        for reaction in self.model.reactions_bidirectional:
            reactants = set(reaction['reactants'])
            products = set(reaction['products'])
            if reaction['reversible']:
                attr_reversible = {'source_arrow_shape': 'diamond', 'target_arrow_shape': 'triangle',
                                   'source_arrow_fill': 'hollow'}
            else:
                attr_reversible = {'source_arrow_shape': 'none', 'target_arrow_shape': 'triangle',
                                      'source_arrow_fill': 'filled'}

            for s in reactants:
                for p in products:
                    self._r_link_species(sp_graph, s, p, **attr_reversible)
        return sp_graph

    @staticmethod
    def _r_link_species(graph, s, r, **attrs):
        """
        Links two nodes in a species graph
        Parameters
        ----------
        s : int
            Source node
        r: int
            Target node
        attrs: dict
            Other attributes for edges

        Returns
        -------

        """

        nodes = ('s{0}'.format(s), 's{0}'.format(r))
        attrs.setdefault('arrowhead', 'normal')
        graph.add_edge(*nodes, **attrs)

    def sp_rxns_bidirectional_graph(self):
        """
        Creates a bipartite nx.DiGraph graph where one set of nodes is the model species
        and the other set is the model bidirectional reactions.

        Returns
        -------
        A :class: nx.Digraph graph that has the information for the visualization of the model
        """

        graph = OrderedGraph(name=self.model.name, graph={'rankdir':'LR'})
        ic_species = [cp for cp, parameter in self.model.initial_conditions]
        for i, cp in enumerate(self.model.species):
            species_node = 's%d' % i
            slabel = hf.parse_name(cp)
            color = "#2b913a"
            # color species with an initial condition differently
            if len([s for s in ic_species if s.is_equivalent_to(cp)]):
                color = "#aaffff"
            graph.add_node(species_node,
                           label=slabel,
                           shape="ellipse",
                           background_color=color,
                           NodeType='species',
                           spInitial=self._sp_initial(cp),
                           bipartite=0)
        for j, reaction in enumerate(self.model.reactions_bidirectional):
            reaction_node = 'r%d' % j
            rule = self.model.rules.get(reaction['rule'][0])
            graph.add_node(reaction_node,
                           label=reaction_node,
                           shape="roundrectangle",
                           background_color="#d3d3d3",
                           NodeType='reaction',
                           kf=rule.rate_forward.value,
                           kr=rule.rate_reverse.value if rule.rate_reverse else 'None',
                           bipartite=1)
            reactants = set(reaction['reactants'])
            products = set(reaction['products'])
            modifiers = reactants & products
            reactants = reactants - modifiers
            products = products - modifiers
            attr_reversible = {'source_arrow_shape': 'triangle', 'target_arrow_shape': 'triangle',
                               'source_arrow_fill': 'hollow'} if reaction['reversible'] \
                                else {'source_arrow_shape': 'none', 'target_arrow_shape': 'triangle', 'source_arrow_fill': 'filled'}
            for s in reactants:
                self._r_link_bipartite(graph, s, j, **attr_reversible)
            for s in products:
                self._r_link_bipartite(graph, s, j, _flip=True, **attr_reversible)
            for s in modifiers:
                attr_modifiers = {'target_arrow_shape':'diamond'}
                self._r_link_bipartite(graph, s, j, **attr_modifiers)
        return graph

    def sp_rxns_graph(self):
        """
        Creates a bipartite nx.DiGraph graph where one set of nodes is the model species
        and the other set is the model unidirectional reactions.

        Returns
        -------
        A :class: nx.Digraph graph that has the information for the visualization of the model
        """
        graph = OrderedGraph(name=self.model.name, graph={'rankdir': 'LR'})
        ic_species = [cp for cp, parameter in self.model.initial_conditions]
        for i, cp in enumerate(self.model.species):
            species_node = 's%d' % i
            slabel = hf.parse_name(self.model.species[i])
            color = "#2b913a"
            # color species with an initial condition differently
            if len([s for s in ic_species if s.is_equivalent_to(cp)]):
                color = "#aaffff"
            graph.add_node(species_node,
                           label=slabel,
                           shape="ellipse",
                           background_color=color,
                           NodeType='species',
                           spInitial=self._sp_initial(cp),
                           bipartite=0)
        for i, reaction in enumerate(self.model.reactions):
            reaction_node = 'r%d' % i
            rule = self.model.rules.get(reaction['rule'][0])
            graph.add_node(reaction_node,
                           label=reaction_node,
                           shape="roundrectangle",
                           background_color="#d3d3d3",
                           NodeType='reaction',
                           kf=rule.rate_forward.value if not reaction['reverse'][0] else 'None',
                           kr=rule.rate_reverse.value if reaction['reverse'][0] else 'None',
                           bipartite=1)
            reactants = set(reaction['reactants'])
            products = set(reaction['products'])
            modifiers = reactants & products
            reactants = reactants - modifiers
            products = products - modifiers
            attr_edges = {'source_arrow_shape': 'none', 'target_arrow_shape': 'triangle', 'source_arrow_fill': 'filled'}
            for s in reactants:
                self._r_link_bipartite(graph, s, i, **attr_edges)
            for s in products:
                self._r_link_bipartite(graph, s, i, _flip=True, **attr_edges)
            for s in modifiers:
                self._r_link_bipartite(graph, s, i, **attr_edges)
        return graph

    def rules_graph(self):
        """
        Creates a bipartite nx.DiGraph graph where one set of nodes is the model species
        and the other set is the model rules.

        Returns
        -------
        A :class: nx.Digraph graph that has the information for the visualization of the model
        """
        graph = self.sp_rxns_graph()
        # Merge reactions into the rules that they come from
        nodes2merge = self.merge_reactions2rules()
        node_attrs = {'shape': 'roundrectangle', 'background_color': '#ff4c4c',
                      'NodeType': 'rule', 'bipartite': 1}
        for rule_info, rxns in nodes2merge.items():
            rule = self.model.rules.get(rule_info[0])
            node_attrs['kf'] = rule.rate_forward.value
            node_attrs['kr'] = rule.rate_reverse.value if rule.rate_reverse else 'None'
            node_attrs['label'] = rule_info[0]
            node_attrs['index'] = 'rule' + str(rule_info[1])
            self.merge_nodes(graph, rxns, rule_info[0], **node_attrs)
        return graph

    def projected_graph(self, graph, project_to='species'):
        """
        Project a bipartite graph into one of the sets of nodes
        Parameters
        ----------
        graph: nx.DiGraph
            a networkx bipartite graph
        project_to: str
            One of the following options: `species_reactions`, `species_rules`,
            `reactions`, `rules`

        Returns
        -------
        a nx.DiGraph
        """
        if project_to == 'species_reactions' or project_to == 'species_rules':
            nodes = {n for n, d in graph.nodes(data=True) if d['bipartite']==0}
        elif project_to == 'reactions' or project_to == 'rules':
            nodes = {n for n, d in graph.nodes(data=True) if d['bipartite']==1}
        else:
            raise ValueError('Projection not valid')
        projected_graph = bipartite.projected_graph(graph, nodes)
        projected_graph = self.graph_merged_pair_edges(projected_graph)

        return projected_graph

    @staticmethod
    def graph_merged_pair_edges(graph):
        """
        Merges pair of edges that are reversed
        Parameters
        ----------
        graph: nx.DiGraph
            The networkx directed graph whose pairs of edges ((u, v), (v, u)) are going to be merged
        Returns
        -------
        A :class: nx.Digraph graph that has the information for the visualization of the model
        """
        edges_to_delete = []
        edges_attributes = {}
        for edge in graph.edges():
            if edge in edges_to_delete:
                continue
            if graph.has_edge(*edge[::-1]):
                attr_reversible = {'source_arrow_shape': 'triangle', 'target_arrow_shape': 'triangle',
                               'source_arrow_fill': 'hollow'}
                edges_attributes[edge] = attr_reversible
                edges_to_delete.append(edge[::-1])
            else:
                attr_irreversible = {'source_arrow_shape': 'none', 'target_arrow_shape': 'triangle',
                                     'source_arrow_fill': 'filled'}
                edges_attributes[edge] = attr_irreversible
        graph.remove_edges_from(edges_to_delete)
        nx.set_edge_attributes(graph, edges_attributes)
        return graph

    def merge_reactions2rules(self):
        """
        Merges the model reactions into each of the rules from which the reactions come form.
        Returns
        -------

        """
        rxn_per_rule = {}
        for r_idx, rule in enumerate(self.model.rules):
            rxns = []
            for i, rxn in enumerate(self.model.reactions):
                if rxn['rule'][0] == rule.name:
                    rxns.append('r{0}'.format(i))
            rxn_per_rule[(rule.name, r_idx)] = rxns
        return rxn_per_rule

    @staticmethod
    def _r_link_bipartite(graph, s, r, **attrs):
        """
        Links two nodes in a species graph
        Parameters
        ----------
        s : int
            Source node
        r: int
            Target node
        attrs: dict
            Other attributes for edges

        Returns
        -------

        """

        nodes = ('s{0}'.format(s), 'r{0}'.format(r))
        if attrs.get('_flip'):
            del attrs['_flip']
            nodes = nodes[::-1]
        attrs.setdefault('arrowhead', 'normal')
        graph.add_edge(*nodes, **attrs)


def graph_to_json(sp_graph, layout=None, path=''):
    """

    Parameters
    ----------
    sp_graph : nx.Digraph graph
        A graph to be converted into cytoscapejs json format
    layout: str or dict
        Name of the layout algorithm to use for the visualization
    path: str
        Path to save the file

    Returns
    -------
    A Dictionary Object that can be converted into Cytoscape.js JSON
    """
    data = from_networkx(sp_graph, layout=layout, scale=1)
    if path:
        with open(path + 'data.json', 'w') as outfile:
            json.dump(data, outfile)
    return data


def dot_layout(sp_graph):
    """

    Parameters
    ----------
    sp_graph : nx.Digraph graph
        Graph to layout

    Returns
    -------
    An OrderedDict containing the node position according to the dot layout
    """

    pos = nx.nx_pydot.graphviz_layout(sp_graph, prog='dot')
    ordered_pos = collections.OrderedDict((node, pos[node]) for node in sp_graph.nodes())
    return ordered_pos



import ipywidgets as widgets
from pyvipr.model_simresult_to_json import data_to_json
from ._version import __frontend_version__

from traitlets import Any, Unicode, Int, observe


@widgets.register
class pysbViz(widgets.DOMWidget):
    """Cytoscape.js widget for simple network visualization."""
    _view_name = Unicode('CytoscapeView').tag(sync=True)
    _model_name = Unicode('CytoscapeModel').tag(sync=True)
    _view_module = Unicode('pyvipr').tag(sync=True)
    _model_module = Unicode('pyvipr').tag(sync=True)
    _view_module_version = Unicode(__frontend_version__).tag(sync=True)
    _model_module_version = Unicode(__frontend_version__).tag(sync=True)

     # Cytoscape options
    data = Any().tag(sync=True, to_json=data_to_json)
    type_of_viz = Unicode('species_view').tag(sync=True, o=True)
    visual_style = Any().tag(sync=True, o=True)
    format = Unicode('cyjs').tag(sync=True, o=True)
    layout_name = Unicode().tag(sync=True, o=True)
    background = Unicode('#FFFFFF').tag(sync=True, o=True)
    random_state = Int(default_value=None, allow_none=True)  # This is necessary only for viz with communities
    process = Unicode('consumption').tag(sync=True, o=True)   # This is necessary only for dynamic visualization
    sim_idx = Int(0).tag(sync=True, o=True)

    @observe('process')
    def _observe_process(self, change):
        self.send_state('data')

    @observe('sim_idx')
    def _observe_sim_idx(self, change):
        self.send_state('data')



"""
Overview
========

PySB implementations of the extrinsic apoptosis reaction model version 1.0
(EARM 1.0) originally published in [Albeck2008]_.

This file contains functions that implement the extrinsic pathway in three
modules:

- Receptor ligation to Bid cleavage (:py:func:`rec_to_bid`)
- Mitochondrial Outer Membrane Permeabilization (MOMP, see below)
- Pore transport to effector caspase activation and PARP cleavage
  (:py:func:`pore_to_parp`).

For the (MOMP) segment there are five variants, which correspond to the five
models described in Figure 11 of [Albeck2008]_:

- "Minimal Model" (Figure 11b, :py:func:`albeck_11b`)
- "Model B + Bax multimerization" (Figure 11c, :py:func:`albeck_11c`)
- "Model C + mitochondrial transport" (Figure 11d, :py:func:`albeck_11d`)
- "Current model" (Figure 11e, :py:func:`albeck_11e`)
- "Current model + cooperativity" (Figure 11f, :py:func:`albeck_11f`)
"""

from pysb import *
from pysb.util import alias_model_components
from pyvipr.pysb_models.shared import *
from pysb.macros import equilibrate

# Default forward, reverse, and catalytic rates:

KF = 1e-6
KR = 1e-3
KC = 1

# Monomer declarations
# ====================

def ligand_to_c8_monomers():
    """ Declares ligand, receptor, DISC, Flip, Bar and Caspase 8.

    'bf' is the site to be used for all binding reactions.

    The 'state' site denotes various localization and/or activity states of a
    Monomer, with 'C' denoting cytoplasmic localization and 'M' mitochondrial
    localization.
    """

    Monomer('L', ['bf']) # Ligand
    Monomer('R', ['bf']) # Receptor
    Monomer('DISC', ['bf']) # Death-Inducing Signaling Complex
    Monomer('flip', ['bf'])
    # Caspase 8, states: pro, Active
    Monomer('C8', ['bf', 'state'], {'state':['pro', 'A']})
    Monomer('BAR', ['bf'])

    alias_model_components()

    # == Annotations
    Annotation(L, 'http://identifiers.org/uniprot/P50591')
    Annotation(R, 'http://identifiers.org/uniprot/O14763')
    Annotation(DISC, 'http://identifiers.org/obo.go/GO:0031264')
    Annotation(flip, 'http://identifiers.org/uniprot/O15519')
    Annotation(C8, 'http://identifiers.org/uniprot/Q14790')
    Annotation(BAR, 'http://identifiers.org/uniprot/Q9NZS9')

def momp_monomers():
    """Declare the monomers used in the Albeck MOMP modules."""

    # == Activators
    # Bid, states: Untruncated, Truncated, truncated and Mitochondrial
    Monomer('Bid', ['bf', 'state'], {'state':['U', 'T', 'M']})

    # == Effectors
    # Bax, states: Cytoplasmic, Mitochondrial, Active
    # sites 's1' and 's2' are used for pore formation
    Monomer('Bax', ['bf', 's1', 's2', 'state'], {'state':['C', 'M', 'A']})

    # == Anti-Apoptotics
    Monomer('Bcl2', ['bf'])

    # == Cytochrome C and Smac
    Monomer('CytoC', ['bf', 'state'], {'state':['M', 'C', 'A']})
    Monomer('Smac', ['bf', 'state'], {'state':['M', 'C', 'A']})

    alias_model_components()

    # == Annotations
    Annotation(Bid, 'http://identifiers.org/uniprot/P55957')
    Annotation(Bax, 'http://identifiers.org/uniprot/Q07812')
    Annotation(Bcl2, 'http://identifiers.org/uniprot/P10415')
    Annotation(CytoC, 'http://identifiers.org/uniprot/P99999')
    Annotation(Smac, 'http://identifiers.org/uniprot/Q9NR28')

def apaf1_to_parp_monomers():
    """ Declares CytochromeC, Smac, Apaf-1, the Apoptosome, Caspases 3, 6, 9,
    XIAP and PARP.

    The package variable 'bf' specifies the name of the site to be used
    for all binding reactions.

    The 'state' site denotes various localization and/or activity states of a
    Monomer, with 'C' denoting cytoplasmic localization and 'M' mitochondrial
    localization.
    """

    # Apaf-1 and Apoptosome
    Monomer('Apaf', ['bf', 'state'], {'state':['I', 'A']}) # Apaf-1

    # Apoptosome (activated Apaf-1 + caspase 9)
    Monomer('Apop', ['bf'])

    # Csp 3, states: pro, active, ubiquitinated
    Monomer('C3', ['bf', 'state'], {'state':['pro', 'A', 'ub']})

    # Caspase 6, states: pro-, Active
    Monomer('C6', ['bf', 'state'], {'state':['pro', 'A']})

    # Caspase 9
    Monomer('C9', ['bf'])

    # PARP, states: Uncleaved, Cleaved
    Monomer('PARP', ['bf', 'state'], {'state':['U', 'C']})

    # X-linked Inhibitor of Apoptosis Protein
    Monomer('XIAP', ['bf'])

    alias_model_components()

    # == Annotations
    Annotation(Apaf, 'http://identifiers.org/uniprot/O14727')
    Annotation(Apop, 'http://identifiers.org/obo.go/GO:0043293')
    Annotation(C3, 'http://identifiers.org/uniprot/P42574')
    Annotation(C6, 'http://identifiers.org/uniprot/P55212')
    Annotation(C9, 'http://identifiers.org/uniprot/P55211')
    Annotation(PARP, 'http://identifiers.org/uniprot/P09874')
    Annotation(XIAP, 'http://identifiers.org/uniprot/P98170')

def all_monomers():
    """Shorthand for calling ligand_to_c8, momp, and apaf1_to_parp macros.

    Internally calls the macros ligand_to_c8_monomers(), momp_monomers(), and
    apaf1_to_parp_monomers() to instantiate the monomers for each portion of the
    pathway.
    """

    ligand_to_c8_monomers()
    momp_monomers()
    apaf1_to_parp_monomers()

# Extrinsic apoptosis module implementations
# ==========================================
#
# These functions implement the upstream (:py:func:`rec_to_bid`) and downstream
# (:py:func:`pore_to_parp`) elements of the extrinsic apoptosis pathway.

def rec_to_bid():
    """Defines the interactions from ligand (e.g. TRAIL) binding to Bid
    activation as per EARM 1.0.

    Uses L, R, DISC, flip, C8, BAR, and Bid monomers and their
    associated parameters to generate the rules that describe Ligand/Receptor
    binding, DISC formation, Caspase-8 activation and
    inhibition by flip and BAR as originally specified in EARM 1.0.

    Declares initial conditions for ligand, receptor, Flip, C8, and Bar.
    """

    # Declare initial conditions for ligand, receptor, Flip, C8, and Bar.
    Parameter('L_0',       3000) # 3000 Ligand corresponds to 50 ng/ml SK-TRAIL
    Parameter('R_0'     ,   200) # 200 TRAIL receptor
    Parameter('flip_0'  , 1.0e2) # Flip 1.0e2
    Parameter('C8_0'    , 2.0e4) # procaspase-8 2.0e4
    Parameter('BAR_0'   , 1.0e3) # Bifunctional apoptosis regulator 1.0e3

    # Needed to recognize the monomer and parameter names in the present scope
    alias_model_components()

    Initial(L(bf=None), L_0)
    Initial(R(bf=None), R_0)
    Initial(flip(bf=None), flip_0)
    Initial(C8(bf=None, state='pro'), C8_0)
    Initial(BAR(bf=None), BAR_0)

    # =====================
    # tBID Activation Rules
    # ---------------------
    #        L + R <--> L:R --> DISC
    #        pC8 + DISC <--> DISC:pC8 --> C8 + DISC
    #        Bid + C8 <--> Bid:C8 --> tBid + C8
    # ---------------------
    catalyze_convert(L(), R(), DISC(bf=None ), [4e-7, KR, 1e-5])
    catalyze(DISC(), C8(state='pro'), C8(state='A'), [KF, KR, KC])
    catalyze(C8(state='A'), Bid(state='U'), Bid(state='T'), [KF, KR, KC])
    # ---------------------
    # Inhibition Rules
    # ---------------------
    #        flip + DISC <-->  flip:DISC
    #        C8 + BAR <--> BAR:C8
    # ---------------------
    bind(DISC(), flip(), [KF, KR])
    bind(BAR(), C8(state='A'), [KF, KR])

def pore_to_parp():
    """Defines what happens after the pore is activated and Cytochrome C and
    Smac are released.

    Uses CytoC, Smac, Apaf, Apop, C3, C6, C8, C9, PARP, XIAP monomers and their
    associated parameters to generate the rules that describe apoptosome
    formation, XIAP inhibition, activation of caspases (including
    caspase-6-mediated feedback), and cleavage of effector caspase substrates
    as specified in EARM 1.0.

    Declares initial conditions for CytoC, Smac, Apaf-1, Apoptosome, caspases
    3, 6, and 9, XIAP, and PARP.
    """

    # Declare initial conditions:

    Parameter('Apaf_0'  , 1.0e5) # Apaf-1
    Parameter('C3_0'    , 1.0e4) # procaspase-3 (pro-C3)
    Parameter('C6_0'    , 1.0e4) # procaspase-6 (pro-C6)
    Parameter('C9_0'    , 1.0e5) # procaspase-9 (pro-C9)
    Parameter('XIAP_0'  , 1.0e5) # X-linked inhibitor of apoptosis protein
    Parameter('PARP_0'  , 1.0e6) # C3* substrate

    alias_model_components()

    Initial(Apaf(bf=None, state='I'), Apaf_0)
    Initial(C3(bf=None, state='pro'), C3_0)
    Initial(C6(bf=None, state='pro'), C6_0)
    Initial(C9(bf=None), C9_0)
    Initial(PARP(bf=None, state='U'), PARP_0)
    Initial(XIAP(bf=None), XIAP_0)

    # CytoC and Smac activation after release
    # --------------------------------------

    equilibrate(Smac(bf=None, state='C'), Smac(bf=None, state='A'),
                          transloc_rates)

    equilibrate(CytoC(bf=None, state='C'), CytoC(bf=None, state='A'),
                          transloc_rates)

    # Apoptosome formation
    # --------------------
    #   Apaf + cCytoC <-->  Apaf:cCytoC --> aApaf + cCytoC
    #   aApaf + pC9 <-->  Apop
    #   Apop + pC3 <-->  Apop:pC3 --> Apop + C3

    catalyze(CytoC(state='A'), Apaf(state='I'), Apaf(state='A'), [5e-7, KR, KC])
    one_step_conv(Apaf(state='A'), C9(), Apop(bf=None), [5e-8, KR])
    catalyze(Apop(), C3(state='pro'), C3(bf=None, state='A'), [5e-9, KR, KC])

    # Apoptosome-related inhibitors
    # -----------------------------
    #   Apop + XIAP <-->  Apop:XIAP
    #   cSmac + XIAP <-->  cSmac:XIAP

    bind(Apop(), XIAP(), [2e-6, KR])
    bind(Smac(state='A'), XIAP(), [7e-6, KR])

    # Caspase reactions
    # -----------------
    # Includes effectors, inhibitors, and feedback initiators:
    #
    #   pC3 + C8 <--> pC3:C8 --> C3 + C8 CSPS
    #   pC6 + C3 <--> pC6:C3 --> C6 + C3 CSPS
    #   XIAP + C3 <--> XIAP:C3 --> XIAP + C3_U CSPS
    #   PARP + C3 <--> PARP:C3 --> CPARP + C3 CSPS
    #   pC8 + C6 <--> pC8:C6 --> C8 + C6 CSPS
    catalyze(C8(state='A'), C3(state='pro'), C3(state='A'), [1e-7, KR, KC])
    catalyze(XIAP(), C3(state='A'), C3(state = 'ub'), [2e-6, KR, 1e-1])
    catalyze(C3(state='A'), PARP(state='U'), PARP(state='C'), [KF, 1e-2, KC])
    catalyze(C3(state='A'), C6(state='pro'), C6(state='A'), [KF, KR, KC])
    catalyze(C6(state='A'), C8(state='pro'), C8(state='A'), [3e-8, KR, KC])

# MOMP module implementations
# ===========================

# Motifs
# ------

# Because several of the models in [Albeck2008]_ overlap, some mechanistic
# aspects have been refactored into the following "motifs", implemented as
# functions:

def Bax_tetramerizes(bax_active_state='A', rate_scaling_factor=1):
    """Creates rules for the rxns Bax + Bax <> Bax2, and Bax2 + Bax2 <> Bax4.

    Parameters
    ----------
    bax_active_state : string: 'A' or 'M'
        The state value that should be assigned to the site "state" for
        dimerization and tetramerization to occur.
    rate_scaling_factor : number
        A scaling factor applied to the forward rate constants for dimerization
        and tetramerization.
    """

    active_unbound = {'state': bax_active_state, 'bf': None}
    active_bax_monomer = Bax(s1=None, s2=None, **active_unbound)
    bax2 =(Bax(s1=1, s2=None, **active_unbound) %
           Bax(s1=None, s2=1, **active_unbound))
    bax4 =(Bax(s1=1, s2=4, **active_unbound) %
           Bax(s1=2, s2=1, **active_unbound) %
           Bax(s1=3, s2=2, **active_unbound) %
           Bax(s1=4, s2=3, **active_unbound))
    Rule('Bax_dimerization', active_bax_monomer + active_bax_monomer | bax2,
         Parameter('Bax_dimerization_kf', KF*rate_scaling_factor),
         Parameter('Bax_dimerization_kr', KR))
    # Notes on the parameter values used below:
    #  - The factor 2 is applied to the forward tetramerization rate because
    #    BNG (correctly) divides the provided forward rate constant by 1/2 to
    #    account for the fact that Bax2 + Bax2 is a homodimerization reaction,
    #    and hence the effective rate is half that of an analogous
    #    heterodimerization reaction. However, Albeck et al. used the same
    #    default rate constant of 1e-6 for this reaction as well, therefore it
    #    must be multiplied by 2 in order to match the original model
    #  - BNG apparently applies a scaling factor of 2 to the reverse reaction
    #    rate, for reasons we do not entirely understand. The factor of 0.5 is
    #    applied here to make the rate match the original Albeck ODEs.
    Rule('Bax_tetramerization', bax2 + bax2 | bax4,
         Parameter('Bax_tetramerization_kf', 2*KF*rate_scaling_factor),
         Parameter('Bax_tetramerization_kr', 0.5*KR))

def Bcl2_binds_Bax1_Bax2_and_Bax4(bax_active_state='A', rate_scaling_factor=1):
    """Creates rules for binding of Bcl2 to Bax monomers and oligomers.

    Parameters
    ----------
    bax_active_state : string: 'A' or 'M'
        The state value that should be assigned to the site "state" for
        the Bax subunits in the pore.
    rate_scaling_factor : number
        A scaling factor applied to the forward rate constants for binding
        between Bax (monomers, oligomers) and Bcl2.
    """

    bind(Bax(state=bax_active_state, s1=None, s2=None), Bcl2,
         [KF*rate_scaling_factor, KR])
    pore_bind(Bax(state=bax_active_state), 's1', 's2', 'bf', 2, Bcl2, 'bf',
         [KF*rate_scaling_factor, KR])
    pore_bind(Bax(state=bax_active_state), 's1', 's2', 'bf', 4, Bcl2, 'bf',
         [KF*rate_scaling_factor, KR])


# Modules
# -------

def albeck_11b(do_pore_transport=True):
    """Minimal MOMP model shown in Figure 11b.

    Features:
        - Bid activates Bax
        - Active Bax is inhibited by Bcl2
        - Free active Bax binds to and transports Smac to the cytosol
    """

    alias_model_components()

    # Set initial conditions
    Initial(Bid(state='U', bf=None), Parameter('Bid_0', 1e5))
    Initial(Bax(bf=None, **inactive_monomer), Parameter('Bax_0', 1e5))
    Initial(Bcl2(bf=None), Parameter('Bcl2_0', 2e4))

    # MOMP Mechanism
    catalyze(Bid(state='T'), Bax(inactive_monomer), Bax(active_monomer),
             [1e-7, KR, KC])
    bind(Bax(active_monomer), Bcl2, [KF, KR])

    # Transport of Smac and Cytochrome C
    if do_pore_transport:
        Initial(Smac(state='M', bf=None), Parameter('Smac_0', 1e6))
        Initial(CytoC(state='M', bf=None), Parameter('CytoC_0', 1e6))
        catalyze(Bax(state='A'), Smac(state='M'), Smac(state='C'),
            [KF, KR, 10])
        catalyze(Bax(state='A'), CytoC(state='M'), CytoC(state='C'),
            [KF, KR, 10])

def albeck_11c(do_pore_transport=True):
    """Model incorporating Bax oligomerization.

    Features:
        - Bid activates Bax
        - Active Bax dimerizes; Bax dimers dimerize to form tetramers
        - Bcl2 binds/inhibits Bax monomers, dimers, and tetramers
        - Bax tetramers bind to and transport Smac to the cytosol
    """

    alias_model_components()
    Initial(Bid(state='U', bf=None), Parameter('Bid_0', 4e4))
    Initial(Bax(bf=None, **inactive_monomer), Parameter('Bax_0', 1e5))
    Initial(Bcl2(bf=None), Parameter('Bcl2_0', 2e4))

    # tBid activates Bax
    catalyze(Bid(state='T'), Bax(inactive_monomer), Bax(active_monomer),
             [1e-7, KR, KC])

    # Bax dimerizes/tetramerizes
    Bax_tetramerizes(bax_active_state='A')

    # Bcl2 inhibits Bax, Bax2, and Bax4
    Bcl2_binds_Bax1_Bax2_and_Bax4(bax_active_state='A')

    if do_pore_transport:
        Initial(Smac(state='M', bf=None), Parameter('Smac_0', 1e6))
        Initial(CytoC(state='M', bf=None), Parameter('CytoC_0', 1e6))
        # NOTE change in KF here from previous model!!!!
        pore_transport(Bax(state='A'), 4, Smac(state='M'), Smac(state='C'),
            [[2*KF, KR, 10]])
        pore_transport(Bax(state='A'), 4, CytoC(state='M'), CytoC(state='C'),
            [[KF, KR, 10]])

def albeck_11d(do_pore_transport=True):
    """Model incorporating mitochondrial transport.

    Features:
        - Bid activates Bax
        - Active Bax translocates to the mitochondria
        - All reactions on the mito membrane have increased association rates
        - Mitochondrial Bax dimerizes; Bax dimers dimerize to form tetramers
        - Bcl2 binds/inhibits Bax monomers, dimers, and tetramers
        - Bax tetramers bind to and transport Smac to the cytosol
    """

    alias_model_components()
    Initial(Bid(state='U', bf=None), Parameter('Bid_0', 4e4))
    Initial(Bax(bf=None, **inactive_monomer), Parameter('Bax_0', 1e5))
    Initial(Bcl2(bf=None), Parameter('Bcl2_0', 2e4))

    # Fractional volume of the mitochondrial membrane compartment
    v = 0.07
    rate_scaling_factor = 1./v

    # tBid activates Bax in the cytosol
    catalyze(Bid(state='T'), Bax(inactive_monomer), Bax(active_monomer),
             [1e-7, KR, KC])

    # Active Bax translocates to the mitochondria
    equilibrate(Bax(bf=None, **active_monomer),
                Bax(bf=None, state='M', s1=None, s2=None),
                [1e-2, 1e-2])

    # Bax dimerizes/tetramerizes
    Bax_tetramerizes(bax_active_state='M',
                     rate_scaling_factor=rate_scaling_factor)

    # Bcl2 inhibits Bax, Bax2, and Bax4
    Bcl2_binds_Bax1_Bax2_and_Bax4(bax_active_state='M',
                                  rate_scaling_factor=rate_scaling_factor)

    if do_pore_transport:
        Initial(Smac(state='M', bf=None), Parameter('Smac_0', 1e6))
        Initial(CytoC(state='M', bf=None), Parameter('CytoC_0', 1e6))
        pore_transport(Bax(state='M'), 4, Smac(state='M'), Smac(state='C'),
            [[rate_scaling_factor*2*KF, KR, 10]])
        pore_transport(Bax(state='M'), 4, CytoC(state='M'), CytoC(state='C'),
            [[KF, KR, 10]])

def albeck_11e(do_pore_transport=True):
    """Model incorporating mitochondrial transport and pore "insertion."

    Features:
        - Bid activates Bax
        - Active Bax translocates to the mitochondria
        - All reactions on the mitochondria have increased association rates
        - Mitochondrial Bax dimerizes; Bax dimers dimerize to form tetramers
        - Bcl2 binds/inhibits Bax monomers, dimers, and tetramers
        - Bax tetramers bind to mitochondrial "sites" and become active pores
        - Active pores bind to and transport Smac to the cytosol
    """
    # Build off of the previous model
    albeck_11d(do_pore_transport=False)

    # Add the "Mito" species, with states "Inactive" and "Active".
    Monomer('Mito', ['bf', 'state'], {'state': ['I', 'A']})
    alias_model_components()
    Initial(Mito(state='I', bf=None), Parameter('Mito_0', 5e5))

    v = 0.07
    rate_scaling_factor = 1./v

    # Add activation of mitochondrial pore sites by Bax4
    pore_bind(Bax(state='M'), 's1', 's2', 'bf', 4, Mito(state='I'), 'bf',
         [KF*rate_scaling_factor, KR])
    Rule('Mito_activation',
         MatchOnce(Bax(state='M', bf=5, s1=1, s2=4) %
                   Bax(state='M', bf=None, s1=2, s2=1) %
                   Bax(state='M', bf=None, s1=3, s2=2) %
                   Bax(state='M', bf=None, s1=4, s2=3) %
                   Mito(state='I', bf=5)) >>
                   Mito(state='A', bf=None),
         Parameter('Mito_activation_kc', KC))

    if do_pore_transport:
        Initial(Smac(state='M', bf=None), Parameter('Smac_0', 1e6))
        Initial(CytoC(state='M', bf=None), Parameter('CytoC_0', 1e6))
        catalyze(Mito(state='A'), Smac(state='M'), Smac(state='C'),
            [rate_scaling_factor*2*KF, KR, 10])
        catalyze(Mito(state='A'), CytoC(state='M'), CytoC(state='C'),
            [rate_scaling_factor*2*KF, KR, 10])

def albeck_11f(do_pore_transport=True):
    """Model as in 11e, but with cooperative assembly of Bax pores.

    Association rate constants for Bax dimerization, tetramerization, and
    insertion are set so that they increase at each step (from 1e-8 to 1e-7 and
    then 1e-6), thereby creating cooperative assembly.

    See also the documentation for albeck_11e().
    """

    albeck_11e(do_pore_transport=do_pore_transport)
    alias_model_components()

    # Set parameter values for cooperative pore formation
    equilibrate_BaxA_to_BaxM_kf.value = 1e-4  # was 1e-2 in 11e
    equilibrate_BaxA_to_BaxM_kr.value = 1e-4  # was 1e-2 in 11e
    Bax_dimerization_kf.value /= 100          # was 1e-6 in 11e
    Bax_tetramerization_kf.value /= 10        # was 1e-6 in 11e





import re
from collections import OrderedDict
from pysb.bng import generate_equations
from pysb.importers.bngl import model_from_bngl
from pysb.importers.sbml import model_from_sbml, model_from_biomodels
import os
import subprocess


def parse_name(spec):
    """
    Function that writes short names of the species to name the nodes.
    It counts how many times a monomer_pattern is present in the complex pattern an its states
    then it takes only the monomer name and its state to write a shorter name to name the nodes.

    Parameters
    ----------
    spec : pysb.ComplexPattern
        Name of species to parse

    Returns
    -------
    Parsed name of species
    """
    m = spec.monomer_patterns
    lis_m = []
    name_counts = OrderedDict()
    parsed_name = ''
    for i in range(len(m)):
        sp_name = str(m[i]).partition('(')[0]
        sp_comp = str(m[i]).partition('** ')[-2:]
        sp_comp = ''.join(sp_comp)
        sp_states = re.findall(r"['\"](.*?)['\"]", str(m[i])) # Matches strings between quotes
        sp_states = [s.lower() for s in sp_states]
        # tmp_2 = re.findall(r"(?<=\').+(?=\')", str(m[i]))
        if not sp_states and not sp_comp:
            lis_m.append(sp_name)
        else:
            sp_states.insert(0, sp_name)
            sp_states.insert(0, sp_comp)
            sp_states.reverse()
            lis_m.append(''.join(sp_states))
    for name in lis_m:
        name_counts[name] = lis_m.count(name)

    for sp, counts in name_counts.items():
        if counts == 1:
            parsed_name += sp + '-'
        else:
            parsed_name += str(counts) + sp + '-'
    return parsed_name[:len(parsed_name) - 1]


def rate_2_interactions(model, rate):
    """
    Obtains the interacting protein from a reaction rate
    Parameters
    ----------
    model : PySB model
    rate : str
    Returns
    -------

    """

    generate_equations(model)
    species_idxs = re.findall('(?<=__s)\d+', rate)
    species_idxs = [int(i) for i in species_idxs]
    if len(species_idxs) == 1:
        interaction = parse_name(model.species[species_idxs[0]])
    else:
        sp_monomers ={sp: model.species[sp].monomer_patterns for sp in species_idxs }
        sorted_intn = sorted(sp_monomers.items(), key=lambda value: len(value[1]))
        interaction = " ".join(parse_name(model.species[mons[0]]) for mons in sorted_intn[:2])
    return interaction


def find_nonimportant_nodes(model):
    """
    This function looks a the bidirectional reactions and finds the nodes that only have one incoming and outgoing
    reaction (edge)

    Parameters
    ----------
    model : pysb.Model
        PySB model to use

    Returns
    -------
    a list of non-important nodes
    """
    if not model.odes:
        generate_equations(model)

    # gets the reactant and product species in the reactions
    rcts_sp = sum([i['reactants'] for i in model.reactions_bidirectional], ())
    pdts_sp = sum([i['products'] for i in model.reactions_bidirectional], ())
    # find the reactants and products that are only used once
    non_imp_rcts = set([x for x in range(len(model.species)) if rcts_sp.count(x) < 2])
    non_imp_pdts = set([x for x in range(len(model.species)) if pdts_sp.count(x) < 2])
    non_imp_nodes = set.intersection(non_imp_pdts, non_imp_rcts)
    passengers = non_imp_nodes
    return passengers


def dispatch_pysb_files(value):
    functions = {'str': _handle_model_files, 'pysb.core.Model': _handle_pysb_model}
    data_type = str(type(value)).split("'")[1]
    result = functions[data_type](value)
    return result


def _handle_model_files(value):
    file_extension = os.path.splitext(value)[1]
    if file_extension == '.bngl':
        model = model_from_bngl(value)
    elif file_extension == '.sbml':
        model = model_from_sbml(value)
    elif value.startswith('BIOMD'):
        model = model_from_biomodels(value)
    # elif file_extension == '.ka':
    #     subprocess.run(['truml', '-k', value])
    #     bngl_model_path = re.sub('ka', 'bngl', value)
    #     model = model_from_bngl(bngl_model_path)
    #     os.remove(bngl_model_path)

    else:
        raise ValueError('Format not supported')
    return model


def _handle_pysb_model(value):
    return value




"""
Overview
========

Three models of MOMP (:py:func:`direct`, :py:func:`indirect`, and
:py:func:`embedded`), each incorporating a larger repertoire of Bcl-2 family
members than previously published models, including:

* One **activator,** Bid.
* Two **sensitizers,** Bad and Noxa.
* Two **effectors,** Bax and Bak.
* Three **anti-apoptotics**, Bcl-2, Bcl-xL, and Mcl-1.

The Models
----------

Note that in each of the three models, interactions between Bcl-2 proteins only
occur at the mitochondrial membrane. The following are brief descriptions of
each model.

* :py:func:`direct`. In this model, tBid directly activates both Bax and Bak;
  the anti-apoptotics bind tBid and the sensitizers (Bad and Noxa) but not
  Bax and Bak.
* :py:func:`indirect`. Bax and Bak are not explicitly activated by tBid, but
  rather are in an equilibrium between inactive and active states. The
  anti-apoptotics bind tBid, sensitizers, and Bax and Bak.
* :py:func:`embedded`. Combines elements of both direct and indirect: tBid
  activates Bax and Bak; the anti-apoptotics bind tBid, sensitizers and Bax and
  Bak. In addition, Bax and Bak are able to auto-activate.

Organization of models into Motifs
----------------------------------

Because the three models share many aspects, the mechanisms that they share have
been written as small "motifs" implemented as subroutines. These are:

* :py:func:`translocate_tBid_Bax_BclxL`
* :py:func:`tBid_activates_Bax_and_Bak`
* :py:func:`tBid_binds_all_anti_apoptotics`
* :py:func:`sensitizers_bind_anti_apoptotics`
* :py:func:`effectors_bind_anti_apoptotics`
* :py:func:`lopez_pore_formation`

The implementation details of these motifs can be seen by examining the
source code.

Monomer and initial declaration functions
-----------------------------------------

The models share the same set of Monomer and initial condition declarations,
which are contained with the following two functions:

* :py:func:`momp_monomers`
* :py:func:`declare_initial_conditions`
"""

# Preliminaries
# =============
#
# We'll need everything from the pysb core, and some macros:

from pysb import *
from pyvipr.pysb_models.shared import *
from pysb.macros import equilibrate
from pysb.util import alias_model_components

# Globals
# -------

# Default rate constants for catalytic activation (fwd, rev, cat):

activation_rates = [        1e-7, 1e-3, 1]

# Shared functions
# ================

# Monomer and initial condition declarations
# ------------------------------------------

def momp_monomers():
    """Declare the monomers for the Bcl-2 family proteins, Cyto c, and Smac.

    Annotation() declarations embedded in this function associate UniProt
    identifiers with  each protein.

    'bf' is the site to be used for all binding reactions (with the
    exception of Bax and Bak, which have additional sites used for
    oligomerization).

    The 'state' site denotes various localization and/or activity states of a
    Monomer, with 'C' denoting cytoplasmic localization and 'M' mitochondrial
    localization. Most Bcl-2 proteins have the potential for both cytoplasmic
    and mitochondrial localization, with the exceptions of Bak and Bcl-2,
    which are apparently constitutively mitochondrial.
    """

    # **Activators**.
    # Bid, states: Untruncated, Truncated, truncated and Mitochondrial
    Monomer('Bid', ['bf', 'state'], {'state':['U', 'T', 'M']})

    # **Effectors**
    # Bax, states: Cytoplasmic, Mitochondrial, Active
    # sites 's1' and 's2' are used for pore formation
    Monomer('Bax', ['bf', 's1', 's2', 'state'], {'state':['C', 'M', 'A']})

    # Bak, states: inactive and Mitochondrial, Active (and mitochondrial)
    # sites 's1' and 's2' are used for pore formation
    Monomer('Bak', ['bf', 's1', 's2', 'state'], {'state':['M', 'A']})

    # **Anti-Apoptotics**
    Monomer('Bcl2', ['bf','state'], {'state':['C', 'M']})
    Monomer('BclxL', ['bf', 'state'], {'state':['C', 'M']})
    Monomer('Mcl1', ['bf', 'state'], {'state':['C', 'M']})

    # **Sensitizers**
    Monomer('Bad', ['bf', 'state'], {'state':['C', 'M']})
    Monomer('Noxa', ['bf', 'state'], {'state': ['C', 'M']})

    # **Cytochrome C and Smac**
    Monomer('CytoC', ['bf', 'state'], {'state':['M', 'C', 'A']})
    Monomer('Smac', ['bf', 'state'], {'state':['M', 'C', 'A']})

    alias_model_components()

    # Annotations
    Annotation(Bid, 'http://identifiers.org/uniprot/P55957')
    Annotation(Bax, 'http://identifiers.org/uniprot/Q07812')
    Annotation(Bak, 'http://identifiers.org/uniprot/Q16611')
    Annotation(Bcl2, 'http://identifiers.org/uniprot/P10415')
    Annotation(BclxL, 'http://identifiers.org/uniprot/Q07817')
    Annotation(Mcl1, 'http://identifiers.org/uniprot/Q07820')
    Annotation(Bad, 'http://identifiers.org/uniprot/Q92934')
    Annotation(Noxa, 'http://identifiers.org/uniprot/Q13794')
    Annotation(CytoC, 'http://identifiers.org/uniprot/P99999')
    Annotation(Smac, 'http://identifiers.org/uniprot/Q9NR28')

def declare_initial_conditions():
    """Declare initial conditions for Bcl-2 family proteins, Cyto c, and Smac.
    """
    Parameter('Bid_0'   , 4.0e4) # Bid 4.0e4
    Parameter('BclxL_0' , 2.0e4) # cytosolic BclxL 2.0e4
    Parameter('Mcl1_0'  , 2.0e4) # Mitochondrial Mcl1 2.0e4
    Parameter('Bcl2_0'  , 2.0e4) # Mitochondrial Bcl2 2.0e4
    Parameter('Bad_0'   , 1e3) # Bad 1e3
    Parameter('Noxa_0'  , 1e3) # Noxa 1e3
    Parameter('CytoC_0' , 5.0e5) # cytochrome c 5.0e5
    Parameter('Smac_0'  , 1.0e5) # Smac 1e5
    Parameter('Bax_0'   , 8e4) # Bax 8e4
    Parameter('Bak_0'   , 2e4) # Bak 2e4

    alias_model_components()

    Initial(Bid(bf=None, state='U'), Bid_0)
    
    Initial(Bax(bf=None, s1=None, s2=None, state='C'), Bax_0)
    Initial(Bak(bf=None, s1=None, s2=None, state='M'), Bak_0)
    
    Initial(Bcl2(bf=None,state='M'), Bcl2_0)
    Initial(BclxL (bf=None, state='M'), BclxL_0)
    Initial(Mcl1(bf=None, state='M'), Mcl1_0)
    
    Initial(Bad(bf=None, state='M'), Bad_0)
    Initial(Noxa(bf=None, state='M'), Noxa_0)
    
    Initial(CytoC(bf=None, state='M'), CytoC_0)
    Initial(Smac(bf=None, state='M'), Smac_0)

# Motifs
# ------

def translocate_tBid_Bax_BclxL():
    """tBid, Bax and BclXL translocate to the mitochondrial membrane."""
    equilibrate(Bid(bf=None, state='T'), Bid(bf=None, state='M'), [1e-1, 1e-3])

    
    #free_Bax = Bax(bf=None, s1=None, s2=None) # Alias for readability
    #equilibrate(free_Bax(state='C'), free_Bax(state='M'),
    #            transloc_rates)

    #equilibrate(BclxL(bf=None, state='C'), BclxL(bf=None, state='M'),
    #            transloc_rates)
    #added by JP, added translocation of sensitizers to membrane
    #equilibrate(Bad(bf=None, state='C'),Bad(bf=None, state='M'),transloc_rates)
    #equilibrate(Noxa(bf=None, state='C'),Noxa(bf=None, state='M'),transloc_rates)


def tBid_activates_Bax_and_Bak():
    """tBid activates Bax and Bak."""
    # added, bid activates bax translocating it to the membrane
    catalyze(Bid(state='M'), Bax(state='C'), Bax(state='M'), activation_rates)
    catalyze(Bid(state='M'), Bax(state='M'), Bax(state='A'), activation_rates)
    catalyze(Bid(state='M'), Bak(state='M'), Bak(state='A'), activation_rates)

def tBid_binds_all_anti_apoptotics():
    """tBid binds and inhibits Bcl2, Mcl1, and Bcl-XL.

    The entries given in the `bind_table` are dissociation constants taken
    from Certo et al. (see ref). Dissociation constants in Certo et al.
    were published as nanomolar binding affinities; here they are converted
    into units of numbers of molecules by multiplying by `N_A` (Avogadro's
    number) and `V` (a default cell volume, specified in :doc:`shared`.

    The default forward rate represents diffusion limited association
    (1e6 Molar^-1 s^-1) and is converted into units of molec^-1 s^-1 by dividing
    by `N_A*V`.

    Certo, M., Del Gaizo Moore, V., Nishino, M., Wei, G., Korsmeyer, S.,
    Armstrong, S. A., & Letai, A. (2006). Mitochondria primed by death signals
    determine cellular addiction to antiapoptotic BCL-2 family members. Cancer
    Cell, 9(5), 351-365. `doi:10.1016/j.ccr.2006.03.027`
    """
    # Doug Green's "MODE 1" inhibition
    bind_table([[             Bcl2(state='M'),  BclxL(state='M'),  Mcl1(state='M')],
                [Bid(state='M'),  66e-9*N_A*V,       12e-9*N_A*V,      10e-9*N_A*V]],
               kf=1e6/(N_A*V))

def sensitizers_bind_anti_apoptotics():
    """Binding of Bad and Noxa to Bcl2, Mcl1, and Bcl-XL.

    See comments on units for :py:func:`tBid_binds_all_anti_apoptotics`.
    """

    bind_table([[                        Bcl2,  BclxL(state='M'),  Mcl1(state='M')],
                [Bad(state='M'),  11e-9*N_A*V,       10e-9*N_A*V,             None],
                [Noxa(state='M'),        None,              None,      19e-9*N_A*V]],
               kf=1e6/(N_A*V))

def effectors_bind_anti_apoptotics():
    """Binding of Bax and Bak to Bcl2, BclxL, and Mcl1.

    Affinities of Bak for Bcl-xL and Mcl-1 are taken from Willis et al.

    Preferential affinity of Bax for Bcl-2 and Bcl-xL were taken from Zhai et
    al.  Bax:Bcl2 and Bax:Bcl-xL affinities were given order of magnitude
    estimates of 10nM.

    See comments on units for :py:func:`tBid_binds_all_anti_apoptotics`.

    Willis, S. N., Chen, L., Dewson, G., Wei, A., Naik, E., Fletcher, J. I.,
    Adams, J. M., et al. (2005). Proapoptotic Bak is sequestered by Mcl-1 and
    Bcl-xL, but not Bcl-2, until displaced by BH3-only proteins. Genes &
    Development, 19(11), 1294-1305. `doi:10.1101/gad.1304105`

    Zhai, D., Jin, C., Huang, Z., Satterthwait, A. C., & Reed, J. C. (2008).
    Differential regulation of Bax and Bak by anti-apoptotic Bcl-2 family
    proteins Bcl-B and Mcl-1. The Journal of biological chemistry, 283(15),
    9580-9586.  `doi:10.1074/jbc.M708426200`
    """

    bind_table([[                            Bcl2,  BclxL(state='M'),Mcl1(state='M')],
                [Bax(active_monomer), 10e-9*N_A*V,       10e-9*N_A*V,         None],
                [Bak(active_monomer),        None,       50e-9*N_A*V,  10e-9*N_A*V]],
               kf=1e6/(N_A*V))

def lopez_pore_formation(do_pore_transport=True):
    """ Pore formation and transport process used by all modules.
    """
    alias_model_components()

    # Rates
    pore_max_size = 4
    pore_rates = [[2.040816e-04,  # 1.0e-6/v**2
                   1e-3]] * (pore_max_size - 1)
    pore_transport_rates = [[2.857143e-5, 1e-3, 10]] # 2e-6 / v?

    # Pore formation by effectors
    assemble_pore_sequential(Bax(bf=None, state='A'), pore_max_size, pore_rates)
    assemble_pore_sequential(Bak(bf=None, state='A'), pore_max_size, pore_rates)

    # CytoC, Smac release
    if do_pore_transport:
        pore_transport(Bax(bf=None, state='A'), pore_max_size, CytoC(state='M'),
                       CytoC(state='C'), pore_transport_rates)
        pore_transport(Bax(bf=None, state='A'), pore_max_size, Smac(state='M'),
                       Smac(state='C'), pore_transport_rates)
        pore_transport(Bak(bf=None, state='A'), pore_max_size, CytoC(state='M'),
                       CytoC(state='C'), pore_transport_rates)
        pore_transport(Bak(bf=None, state='A'), pore_max_size, Smac(state='M'),
                       Smac(state='C'), pore_transport_rates)

# MOMP model implementations
# ==========================

def embedded(do_pore_transport=True):
    """ Direct and indirect modes of action, occurring at the membrane.
    """
    alias_model_components()

    declare_initial_conditions()

    translocate_tBid_Bax_BclxL()

    tBid_activates_Bax_and_Bak()

    # Autoactivation: Bax and Bak activate their own kind, but only when
    # free (i.e. not part of a pore complex)
    catalyze(Bax(active_monomer), Bax(state='M'), Bax(state='A'),
             activation_rates)
    catalyze(Bak(active_monomer), Bak(state='M'), Bak(state='A'),
             activation_rates)

    # Anti-apoptotics bind activator tBid
    # Doug Green's "MODE 1" inhibition
    tBid_binds_all_anti_apoptotics()

    # Anti-apoptotics bind activated effectors
    # Doug Green's "MODE 2" inhibition
    effectors_bind_anti_apoptotics()

    sensitizers_bind_anti_apoptotics()
    
    # Bax and Bak form pores by sequential addition and transport CytoC/Smac
    lopez_pore_formation(do_pore_transport=do_pore_transport)

def indirect(do_pore_transport=True):
    """Bax and Bak spontaneously form pores without activation.
       The "activator" tBid binds all of the anti-apoptotics.
    """
    alias_model_components()

    declare_initial_conditions()

    translocate_tBid_Bax_BclxL()

    # Bax and Bak spontaneously become activated
    free_Bax = Bax(bf=None, s1=None, s2=None) # Alias
    free_Bak = Bak(bf=None, s1=None, s2=None) # Alias
    free_Bax = Bax(bf=None, s1=None, s2=None) # Alias for readability
    equilibrate(free_Bax(state='C'), free_Bax(state='M'),
                transloc_rates)
    equilibrate(free_Bax(state='M'), free_Bax(state='A'), transloc_rates)
    equilibrate(free_Bak(state='M'), free_Bak(state='A'), transloc_rates)
    
    # Anti-apoptotics bind activator tBid
    # Doug Green's "MODE 1" inhibition
    tBid_binds_all_anti_apoptotics()

    # Anti-apoptotics bind activated effectors
    # Doug Green's "MODE 2" inhibition
    effectors_bind_anti_apoptotics()

    sensitizers_bind_anti_apoptotics()

    # Bax and Bak form pores by sequential addition
    lopez_pore_formation(do_pore_transport=do_pore_transport)

def direct(do_pore_transport=True):
    """Anti-apoptotics prevent BH3-onlies from activating Bax and Bak.

    Bax and Bak require activation to be able to form pores.
    The anti-apoptotics don't inhibit activated Bax and Bak; their only role
    is to bind BH3-onlies.
    """

    alias_model_components()
    declare_initial_conditions()

    translocate_tBid_Bax_BclxL()


    tBid_activates_Bax_and_Bak()

    # Anti-apoptotics bind activator tBid
    # Doug Green's "MODE 1" inhibition
    tBid_binds_all_anti_apoptotics()

    sensitizers_bind_anti_apoptotics()
    
    # Bax and Bak form pores by sequential addition
    lopez_pore_formation(do_pore_transport=do_pore_transport)




from pyvipr.static_viz import graph_to_json, dot_layout


class NetworkViz(object):
    def __init__(self, network):
        self.network = network

    def network_static_view(self):
        self.network.graph['name'] = ''
        d_layout = dot_layout(self.network)
        json = graph_to_json(self.network, layout=d_layout)
        return json

    def dynamic_network_view(self):
        # This is necessary because the javascript part chooses to render
        # a static or dynamic visualization depending on the name
        # of the function.
        g_layout = dot_layout(self.network)
        data = graph_to_json(sp_graph=self.network, layout=g_layout)
        return data



from pysb import Model, Monomer, Parameter, Compartment, Rule, Initial

Model()

Parameter('vol_CYT', 1)
Parameter('vol_MOM', 1)

Compartment('CYT', parent=None, dimension=3, size=vol_CYT)
Compartment('MOM', parent=CYT, dimension=2, size=vol_MOM)

Monomer('tBid', ['bf'])
Monomer('Bax', ['bf', 's1', 's2', 'state'], {'state': ['I', 'A']})
Monomer('Bcl2', ['bf'])

Parameter('Bid_0', 40000.0)
Parameter('Bcl2_0', 20000.0)
Parameter('Bax_0', 80000.0)

Parameter('equilibrate_BidT_to_BidM_kf', 0.1)
Parameter('equilibrate_BidT_to_BidM_kr', 0.001)
Parameter('equilibrate_BaxC_to_BaxM_kf', 0.01)
Parameter('equilibrate_BaxC_to_BaxM_kr', 0.01)
Parameter('bind_BidM_Bcl2_kf', 1e-06)
Parameter('bind_BidM_Bcl2_kr', 0.06600000000000002)
Parameter('bind_BidM_BaxM_to_BidMBaxM_kf', 1e-07)
Parameter('bind_BidM_BaxM_to_BidMBaxM_kr', 0.001)
Parameter('catalyze_BidMBaxM_to_BidM_BaxA_kc', 1.0)

Initial(tBid(bf=None) ** CYT, Bid_0)
Initial(Bax(bf=None, s1=None, s2=None, state='I') ** CYT, Bax_0)
Initial(Bcl2(bf=None) ** MOM, Bcl2_0)

Rule('equilibrate_BidT_to_BidM', tBid(bf=None) ** CYT | tBid(bf=None) ** MOM, equilibrate_BidT_to_BidM_kf, equilibrate_BidT_to_BidM_kr)
Rule('equilibrate_BaxC_to_BaxM', Bax(bf=None, s1=None, s2=None, state='I') ** CYT | Bax(bf=None, s1=None, s2=None, state='I') ** MOM, equilibrate_BaxC_to_BaxM_kf, equilibrate_BaxC_to_BaxM_kr)
Rule('bind_BidM_BaxM_to_BidMBaxM', tBid(bf=None) ** MOM + Bax(bf=None, s1=None, s2=None, state='I') ** MOM | tBid(bf=1) ** MOM % Bax(bf=1, s1=None, s2=None, state='I') ** MOM, bind_BidM_BaxM_to_BidMBaxM_kf, bind_BidM_BaxM_to_BidMBaxM_kr)
Rule('catalyze_BidMBaxM_to_BidM_BaxA', tBid(bf=1) ** MOM % Bax(bf=1, s1=None, s2=None, state='I') ** MOM >> tBid(bf=None) ** MOM + Bax(bf=None, s1=None, s2=None, state='A') ** MOM, catalyze_BidMBaxM_to_BidM_BaxA_kc)
Rule('bind_BidM_Bcl2', tBid(bf=None) ** MOM + Bcl2(bf=None) ** MOM | tBid(bf=1) ** MOM % Bcl2(bf=1) ** MOM, bind_BidM_Bcl2_kf, bind_BidM_Bcl2_kr)



"""
Data conversion utility for NetworkX
=====================================
Convert cytoscape.js style graphs from/to NetworkX object.
https://networkx.github.io/
"""

import networkx as nx

# Special Keys
ID = 'id'
NAME = 'name'
DATA = 'data'
ELEMENTS = 'elements'

NODES = 'nodes'
EDGES = 'edges'

SOURCE = 'source'
TARGET = 'target'

DEF_SCALE = 100


def __map_table_data(columns, graph_obj):
    data = {}
    for col in columns:
        if col == 0:
            break

        data[col] = graph_obj[col]

    return data


def __create_node(node, node_id):
    new_node = {}
    node_columns = node.keys()
    data = __map_table_data(node_columns, node)
    # Override special keys
    data[ID] = str(node_id)
    data[NAME] = str(node_id)

    if 'position' in node.keys():
        position = node['position']
        new_node['position'] = position

    new_node[DATA] = data
    return new_node


def __build_multi_edge(edge_tuple, g):
    source = edge_tuple[0]
    target = edge_tuple[1]
    key = edge_tuple[2]
    data = edge_tuple[3]

    data['source'] = str(source)
    data['target'] = str(target)
    data['interaction'] = str(key)
    return {DATA: data}


def __build_edge(edge_tuple, g):
    source = edge_tuple[0]
    target = edge_tuple[1]
    data = edge_tuple[2]

    data['source'] = str(source)
    data['target'] = str(target)
    return {DATA: data}


def __build_empty_graph():
    return {
        DATA: {},
        ELEMENTS: {
            NODES: [],
            EDGES: []
        }
    }


def from_networkx(g, layout=None, scale=DEF_SCALE):
    # Dictionary Object to be converted to Cytoscape.js JSON
    cygraph = __build_empty_graph()

    if layout is not None:
        pos = list(map(lambda position:
                  {'x': position[0]*scale, 'y': position[1]*scale},
                  layout.values()))

    nodes = g.nodes()
    if isinstance(g, nx.MultiDiGraph) or isinstance(g, nx.MultiGraph):
        edges = g.edges(data=True, keys=True)
        edge_builder = __build_multi_edge
    else:
        edges = g.edges(data=True)
        edge_builder = __build_edge

    # Map network table data
    cygraph[DATA] = __map_table_data(g.graph.keys(), g.graph)

    for i, node_id in enumerate(nodes):
        new_node = __create_node(g.node[node_id], node_id)
        if layout is not None:
            new_node['position'] = pos[i]

        cygraph['elements']['nodes'].append(new_node)

    for edge in edges:
        cygraph['elements']['edges'].append(edge_builder(edge, g))

    return cygraph


def to_networkx(cyjs, directed=True):
    """
    Convert Cytoscape.js-style JSON object into NetworkX object.
    By default, data will be handles as a directed graph.
    """

    if directed:
        g = nx.MultiDiGraph()
    else:
        g = nx.MultiGraph()

    network_data = cyjs[DATA]
    if network_data is not None:
        for key in network_data.keys():
            g.graph[key] = network_data[key]

    nodes = cyjs[ELEMENTS][NODES]
    edges = cyjs[ELEMENTS][EDGES]

    for node in nodes:
        data = node[DATA]
        g.add_node(data[ID], attr_dict=data)

    for edge in edges:
        data = edge[DATA]
        source = data[SOURCE]
        target = data[TARGET]

        g.add_edge(source, target, attr_dict=data)

    return g


def network_dynamic_data(network, tspan, node_rel=None, node_tip=None, edge_colors=None, edge_sizes=None,
                         edge_tips=None):
    """

    Parameters
    ----------
    network: nx.DiGraph or nx.Graph
    tspan
    node_rel: dict
        A dictionary where the keys are the node ids and the values are
        lists that contain (0-100) values that are represented in a
        pie chart within the node
    node_tip: dict
        A dictionary where the keys are the node ids and the values are
        lists that contain any value that can be accessed
        as a tooltip in the rendered network
    edge_colors: dict
        A dictionary where the keys are the edge ids and the values are
        lists that contain any hexadecimal color value that are
        represented in the edge colors
    edge_sizes: dict
        A dictionary where the keys are the edge ids and the values are
        lists that contain any numerical value that are
        represented in the edge size
    edge_tips: dict
        A dictionary where the keys are the edge ids and the values are
        lists that contain any value that can be accessed
        as a tooltip in the rendered network
-

    """
    node_rel_default = [0] * len(tspan)
    node_tip_default = [0] * len(tspan)
    edge_colors_default = ['#787878'] * len(tspan)
    edge_sizes_default = [5] * len(tspan)
    edge_tips_default = [0] * len(tspan)

    if node_rel is not None:
        node_rel_default = node_rel
    if node_tip is not None:
        node_tip_default = node_tip
    if edge_colors is not None:
        edge_colors_default = edge_colors
    if edge_sizes is not None:
        edge_sizes_default = edge_sizes
    if edge_tips is not None:
        edge_tips_default = edge_tips

    node_edge_properties = [node_rel_default, node_tip_default, edge_colors_default,
                            edge_sizes_default, edge_tips_default]

    if [idx for idx in node_edge_properties if len(idx) != len(tspan)]:
        raise ValueError('All edge and node properties must be '
                         'equal length')

    nx.set_node_attributes(network, node_rel_default, 'rel_value')
    nx.set_node_attributes(network, node_tip_default, 'abs_value')
    nx.set_node_attributes(network, 'ellipse', 'shape')
    nx.set_node_attributes(network, '#2b913a', 'background_color')
    nx.set_edge_attributes(network, edge_colors_default, 'edge_color')
    nx.set_edge_attributes(network, edge_sizes_default, 'edge_size')
    nx.set_edge_attributes(network, edge_tips_default, 'edge_qtip')

    network.graph['name'] = ''
    network.graph['view'] = 'dynamic'
    network.graph['tspan'] = tspan




from .pysbviz import pysbViz

__all__ = [
    'sp_view',
    'sp_comp_view',
    'sp_comm_view',
    'sp_rxns_bidirectional_view',
    'sp_rxns_view',
    'sp_rules_view',
    'sp_rules_fxns_view',
    'sp_rules_mod_view',
    'projected_species_reactions_view',
    'projected_reactions_view',
    'projected_rules_view',
    'projected_species_rules_view',
    'sp_dyn_view',
    'sp_comp_dyn_view',
    'sp_comm_dyn_view',
    'sim_model_dyn_view',
    'nx_graph_view',
    'nx_graph_dyn_view'
]


def sp_view(model, layout_name='preset'):
    """
    Render a visualization of the interactions between the species in a model.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """

    return pysbViz(data=model, type_of_viz='species_view', layout_name=layout_name)


def sp_comp_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of the interactions between the species in a model.
     The species nodes are grouped y the compartments they belong to.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return pysbViz(data=model, type_of_viz='species_compartments_view', layout_name=layout_name)


def sp_comm_view(model, layout_name='klay', random_state=None):
    """
    Render a visualization of the interactions between the species in a model.
     The species nodes are grouped y the communities detected by the Louvain algorithm.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use
    random_state: int
        Random state seed use by the community detection algorithm

    """
    return pysbViz(data=model, type_of_viz='communities_view', random_state=random_state, layout_name=layout_name)


def sp_rxns_bidirectional_view(model, layout_name='preset'):
    """
    Render a visualization of a bipartite graph where one set of nodes
    are the molecular species in the model and the other set are the
    bidirectional reactions.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return pysbViz(data=model, type_of_viz='sp_rxns_bidirectional_view', layout_name=layout_name)


def sp_rxns_view(model, layout_name='preset'):
    """
    Render a visualization of a bipartite graph where one set of nodes
    are the molecular species in the model and the other set are the
    unidirectional reactions.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return pysbViz(data=model, type_of_viz='sp_rxns_view', layout_name=layout_name)


def sp_rules_view(model, layout_name='preset'):
    """
    Render a visualization of a bipartite graph where one set of nodes
    are the molecular species in the model and the other set are the rules.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return pysbViz(data=model, type_of_viz='sp_rules_view', layout_name=layout_name)


def sp_rules_fxns_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of a bipartite graph where one set of nodes
    are the molecular species in the model and the other set are the rules.
    Additionally, rules are grouped by the macros that created them.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return pysbViz(data=model, type_of_viz='sp_rules_functions_view', layout_name=layout_name)


def sp_rules_mod_view(model, layout_name='cose-bilkent'):
    """
    Render a visualization of a bipartite graph where one set of nodes
    are the molecular species in the model and the other set are the rules.
    Additionally, rules are grouped by the modules where they are defined.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return pysbViz(data=model, type_of_viz='sp_rules_modules_view', layout_name=layout_name)


def projected_species_reactions_view(model, layout_name='preset'):
    """
    Render a visualization of the interactions between species in a model.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return pysbViz(data=model, type_of_viz='projected_species_reactions_view', layout_name=layout_name)


def projected_reactions_view(model, layout_name='preset'):
    """
    Render a visualization of the interaction between the reaction in a model

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return pysbViz(data=model, type_of_viz='projected_reactions_view', layout_name=layout_name)


def projected_rules_view(model, layout_name='preset'):
    """
    Render a visualization of the interactions between rules in a model.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return pysbViz(data=model, type_of_viz='projected_rules_view', layout_name=layout_name)


def projected_species_rules_view(model, layout_name='preset'):
    """
    Render a visualization of a bipartite graph where one set of nodes
    are the molecular species in the model and the other set are the rules.

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    layout_name: str
        Layout to use

    """
    return pysbViz(data=model, type_of_viz='projected_species_rules_view', layout_name=layout_name)


def sp_dyn_view(simulation, sim_idx=0, process='consumption', layout_name='preset'):
    """
    Render a dynamic visualization of the simulation

    Parameters
    ----------
    simulation : pysb.SimulationResult
        Simulation result to visualize
    sim_idx : int
        Index of simulation to be visualized
    process : str
        Type of the dynamic visualization, it can be 'consumption' or 'production'
    layout_name : str
        Layout to use

    """
    return pysbViz(data=simulation, type_of_viz='dynamic_view', layout_name=layout_name,
                   process=process, sim_idx=sim_idx)


def sp_comp_dyn_view(simulation, sim_idx=0, process='consumption', layout_name='cose-bilkent'):
    """
    Render a dynamic visualization of the simulation. The species nodes are grouped
    by the compartments they belong to.

    Parameters
    ----------
    simulation: pysb.SimulationResult object
        Simulation result to visualize dynamically
    sim_idx: int
        Index of simulation to be visualized
    process : str
        Type of the dynamic visualization, it can be 'consumption' or 'production'
    layout_name: str
        Layout to use

    """
    return pysbViz(data=simulation, type_of_viz='dynamic_compartments_view', layout_name=layout_name,
                   process=process, sim_idx=sim_idx)


def sp_comm_dyn_view(simulation, sim_idx=0, process='consumption', layout_name='klay', random_state=None):
    """
    Render a dynamic visualization of the simulation. The species nodes are grouped
    by the communities detected by the Louvain algorithm

    Parameters
    ----------
    simulation: pysb.SimulationResult object
        Simulation result to visualize dynamically
    sim_idx: int
        Index of simulation to be visualized
    process : str
        Type of the dynamic visualization, it can be 'consumption' or 'production'
    layout_name: str
        Layout to use
    random_state: int
        Random state seed use by the community detection algorithm

    """
    return pysbViz(data=simulation, type_of_viz='dynamic_communities_view', layout_name=layout_name,
                   random_state=random_state, process=process, sim_idx=sim_idx)


def sim_model_dyn_view(model, tspan, param_values=None, type_of_viz='dynamic_view',
                       process='consumption', layout_name='preset'):
    """
    Render a dynamic visualization of the model using the tspan and param_values
    passed to the function

    Parameters
    ----------
    model: pysb.model or str
        Model to visualize. It can be a pysb model, or the file path to an
        an SBML or BNGL model
    tspan : vector-like, optional
        Time values over which to simulate. The first and last values define
        the time range.
    param_values : vector-like or dict, optional
        Values to use for every parameter in the model. Ordering is
        determined by the order of model.parameters.
        If passed as a dictionary, keys must be parameter names.
        If not specified, parameter values will be taken directly from
        model.parameters.
    type_of_viz: str
        Type of visualization. It can only be `dynamic_view`, `dynamic_compartments_view`
        or `dynamic_communities_view`
    process : str
        Type of the dynamic visualization, it can be 'consumption' or 'production'
    layout_name : str
        Layout name to use

    """
    from pysb.simulator import ScipyOdeSimulator
    from pyvipr.util import dispatch_pysb_files

    model = dispatch_pysb_files(model)
    sim = ScipyOdeSimulator(model, tspan=tspan).run(param_values=param_values)
    return pysbViz(data=sim, type_of_viz=type_of_viz, process=process, layout_name=layout_name)


def nx_graph_view(graph, layout_name='cose'):
    """
    Render a networkx Graph or DiGraph

    Parameters
    ----------
    graph: nx.Graph or nx.DiGraph
        Graph to render
    layout_name: str
        Layout to use

    Returns
    -------

    """
    return pysbViz(data=graph, type_of_viz='network_static_view', layout_name=layout_name)


def nx_graph_dyn_view(graph, tspan, node_rel=None, node_tip=None, edge_colors=None, edge_sizes=None,
                      edge_tips=None, layout_name='cose'):
    """
    Render a dynamic visualization of a networkx graph

    Parameters
    ----------
    graph: nx.DiGraph or nx.Graph
    tspan
    node_rel: dict
        A dictionary where the keys are the node ids and the values are
        lists that contain (0-100) values that are represented in a
        pie chart within the node
    node_tip: dict
        A dictionary where the keys are the node ids and the values are
        lists that contain any value that can be accessed
        as a tooltip in the rendered network
    edge_colors: dict
        A dictionary where the keys are the edge ids and the values are
        lists that contain any hexadecimal color value that are
        represented in the edge colors
    edge_sizes: dict
        A dictionary where the keys are the edge ids and the values are
        lists that contain any numerical value that are
        represented in the edge size
    edge_tips: dict
        A dictionary where the keys are the edge ids and the values are
        lists that contain any value that can be accessed
        as a tooltip in the rendered network
    layout_name: str
        Layout to use

    """
    from pyvipr.util_networkx import network_dynamic_data

    network_dynamic_data(graph, tspan, node_rel, node_tip, edge_colors, edge_sizes,
                         edge_tips)

    return pysbViz(data=graph, type_of_viz='dynamic_network_view', layout_name=layout_name)



"""
Overview
========

This module declares a number of functions and variables that are
used by many of the EARM 2 models. The functions can be divided into the
following four categories:

1. Functions that are specific to the models in EARM 2, but are used by all of
   them. The only macro of this type is

   - :py:func:`observables`

2. Aliases to generalized macros in pysb.macros that provide default values
   for site names or other arguments. Macros of this type include:

   - :py:func:`catalyze`
   - :py:func:`bind`
   - :py:func:`bind_table`
   - :py:func:`assemble_pore_sequential`
   - :py:func:`pore_transport`

3. Macros for mechanisms that appear within the models previously published by
   the research group of Pingping Shen (or the model from [Howells2011]_, which
   is derived from one of Shen's models):

   - :py:func:`assemble_pore_spontaneous`
   - :py:func:`displace`
   - :py:func:`displace_reversibly`

4. Macros for mechanisms that appear within the models described in our
   group's earlier work, specifically the models described in [Albeck2008]_:

   - :py:func:`catalyze_convert`
   - :py:func:`one_step_conv`
   - :py:func:`pore_bind`
"""

# Preliminaries
# =============

# We need the main things from pysb.core plus a few extras:

from pysb import *
from pysb import MonomerPattern, ComplexPattern, ComponentSet
import pysb.macros as macros
from pysb.util import alias_model_components
import functools

# Also pull in Avogadro's number from Scipy:

from scipy.constants import N_A

# Global variables
# ================

# Default constants
# -----------------

# **Default cell volume** to be used when converting between stochastic and
# deterministic rates. The value used is
#
# (1/N_A) * 1e12 L = 1.661e-12 L (1.661 picoliters or 1661 um^3)
#
# is at the lower end of the range for whole cell volume,
# (see `BioNumbers ID 103725
# <http://bionumbers.hms.harvard.edu/bionumber.aspx?&id=103725&ver=14&trm=hela%20cell%20volume>`_).
# but it is slightly above the range given for the volume of HeLa cytoplasm
# (see `BioNumbers ID 106664
# <http://bionumbers.hms.harvard.edu/bionumber.aspx?&id=106664&ver=3&trm=hela%20cell%20volume>`_).
#
# One nice thing about using this value is that since
# 1.661e-12 L * N_A = 1e12 L, it leads to the rule of thumb that
# 1e-9 Molar = 1000 molecules:
#
# [1e-9 Moles/L] = [No. of molecules] / (N_A * vol)
#
# [1e-9 Moles/L] = [No. of molecules] / 1e12 L
#
# [1e-9 Moles/L] * [1e12 L/Moles] = [No. of molecules]
#
# [No. of molecules] = 1e3

V = (1/N_A) * 1e12

# **Default forward and reverse rates** for translocation reactions:

transloc_rates = [1e-2, 1e-2]

# **Rate scaling** for reactions occurring on the mitochondrial membrane.
# `mito_fractional_volume` represents the fractional volume of the
# mitochondrial membrane compartment, so the forward rate constants for
# reactions on the membrane is `1/v`. The approach and the value used is
# adopted from [Albeck2008]_.

mito_fractional_volume = 0.07
rate_scaling_factor = 1./mito_fractional_volume

# Aliases
# -------

# Some useful aliases for typical Bax/Bak states:

active_monomer = {'state':'A', 's1': None, 's2': None}
inactive_monomer = {'state':'C', 's1': None, 's2': None}

# Observables declarations
# ========================

def observables():
    """Declare observables commonly used for the TRAIL pathway.

    Declares truncated (and mitochondrial) Bid, cytosolic (i.e., released)
    Smac, and cleaved PARP.
    """

    alias_model_components()
    # Observables
    # ===========
    Observable('mBid',  Bid(state='M'))
    Observable('aSmac', Smac(state='A'))
    Observable('cPARP', PARP(state='C'))

# Aliases to pysb.macros
# ======================

def catalyze(enz, sub, product, klist):
    """Alias for pysb.macros.catalyze with default binding sites."""

    return macros.catalyze(enz, 'bf', sub, 'bf', product, klist)

def bind(a, b, klist):
    """Alias for pysb.macros.bind with default binding sites."""

    return macros.bind(a, 'bf', b, 'bf', klist)

def bind_table(table, **kwargs):
    """Alias for pysb.macros.bind_table with default binding sites."""

    return macros.bind_table(table, 'bf', 'bf', **kwargs)

def assemble_pore_sequential(subunit, size, klist):
    """Alias for pysb.macros.assemble_pore_sequential with default sites.

    Uses default pore site names as the sites for subunit-subunit binding in
    the pore.
    """

    return macros.assemble_pore_sequential(subunit, 's1', 's2', size, klist)

def pore_transport(subunit, size, csource, cdest, ktable):
    """Alias for pysb.macros.pore_transport with default arguments.

    - Uses the default binding site names for the binding site on the pore
      and on the cargo
    - Uses the default pore site names for subunit-subunit binding
    - Uses only a single size (not a min and max size) for the size of
      transport-competent pores
    """

    return macros.pore_transport(subunit, 's1', 's2', 'bf',
                                 size, size, csource, 'bf', cdest, ktable)

def pore_bind(subunit, size, cargo, klist):
    """Alias for pysb.macros.pore_bind with default arguments.

    - Uses the default binding site names for the binding site on the pore
      and on the cargo
    - Uses the default pore site names for subunit-subunit binding
    """

    return macros.pore_bind(subunit, 's1', 's2', 'bf',
                                 size, cargo, 'bf', klist)


# Macros used by the Shen models
# ==============================

def assemble_pore_spontaneous(subunit, klist):
    """Generate the order-4 assembly reaction 4*Subunit <> Pore."""

    # This is a function that is passed to macros._macro_rule to generate
    # the name for the pore assembly rule. It follows the pattern of,
    # e.g., "BaxA_to_BaxA4" for a Bax pore of size 4.
    def pore_rule_name(rule_expression):
        react_p = rule_expression.reactant_pattern
        mp = react_p.complex_patterns[0].monomer_patterns[0]
        subunit_name = macros._monomer_pattern_label(mp)
        pore_name = mp.monomer.name
        return '%s_to_%s%d' % (subunit_name, mp.monomer.name, 4)

    # Alias for a subunit that is capable of forming a pore
    free_subunit = subunit(s1=None, s2=None)

    # Create the pore formation rule
    macros._macro_rule('spontaneous_pore',
        free_subunit + free_subunit + free_subunit + free_subunit |
        subunit(s1=1, s2=4) % subunit(s1=2, s2=1) % \
        subunit(s1=3, s2=2) % subunit(s1=4, s2=3),
        klist, ['kf', 'kr'], name_func=pore_rule_name)

def displace(lig1, lig2, target, k):
    """Generate unidirectional displacement reaction L1 + L2:T >> L1:T + L2.

    The signature can be remembered with the following formula:
    "lig1 displaces lig2 from target."
    """

    return macros._macro_rule('displace',
         lig1({'bf':None}) + lig2({'bf':1}) % target({'bf':1}) >>
         lig1({'bf':1}) % target({'bf':1}) + lig2({'bf':None}),
         [k], ['kf'])

def displace_reversibly(lig1, lig2, target, klist):
    """Generate reversible displacement reaction L1 + L2:T <> L1:T + L2.

    The signature can be remembered with the following formula:
    "lig1 displaces lig2 from target." The first rate given in
    in klist specifies the forward rate of this reaction; the second
    specifies the reverse rate.
    """

    return macros._macro_rule('displace',
         lig1({'bf':None}) + lig2({'bf':1}) % target({'bf':1}) |
         lig1({'bf':1}) % target({'bf':1}) + lig2({'bf':None}),
         klist, ['fwd_kf', 'rev_kf'])

# Macros used by the Albeck models
# ================================

def catalyze_convert(sub1, sub2, product, klist, site='bf'):
    """Automation of the Sub1 + Sub2 <> Sub1:Sub2 >> Prod two-step reaction.

    Because product is created by the function, it must be fully specified.
    """

    # Make sure that the substrates have the site:
    macros._verify_sites(sub1, site)
    macros._verify_sites(sub2, site)

    components = macros._macro_rule('bind',
                             sub1({site: None}) + sub2({site: None}) |
                             sub1({site: 1}) % sub2({site: 1}),
                             klist[0:2], ['kf', 'kr'])
    components |= macros._macro_rule('convert',
                              sub1({site: 1}) % sub2({site: 1}) >> product,
                              [klist[2]], ['kc'])
    return components

def one_step_conv(sub1, sub2, product, klist, site='bf'):
    """ Bind sub1 and sub2 to form one product: sub1 + sub2 <> product.
    """

    kf, kr = klist

    # Make sure that the substrates have the site:
    macros._verify_sites(sub1, site)
    macros._verify_sites(sub2, site)

    return macros._macro_rule('convert',
                       sub1({site: None}) + sub2({site: None}) | product,
                       klist, ['kf', 'kr'])

def pore_bind(subunit, sp_site1, sp_site2, sc_site, size, cargo, c_site,
              klist):
    """Generate rules to bind a monomer to a circular homomeric pore.

    The pore structure is defined by the `pore_species` macro -- `subunit`
    monomers bind to each other from `sp_site1` to `sp_site2` to form a closed
    ring. The binding reaction takes the form pore + cargo <> pore:cargo.

    Parameters
    ----------
    subunit : Monomer or MonomerPattern
        Subunit of which the pore is composed.
    sp_site1, sp_site2 : string
        Names of the sites where one copy of `subunit` binds to the next.
    sc_site : string
        Name of the site on `subunit` where it binds to the cargo `cargo`.
    size : integer
        Number of subunits in the pore at which binding will occur.
    cargo : Monomer or MonomerPattern
        Cargo that binds to the pore complex.
    c_site : string
        Name of the site on `cargo` where it binds to `subunit`.
    klist : list of Parameters or numbers
        List containing forward and reverse rate constants for the binding
        reaction (in that order). Rate constants should either be both Parameter
        objects or both numbers. If Parameters are passed, they will be used
        directly in the generated Rules. If numbers are passed, Parameters
        will be created with automatically generated names based on <TODO>
        and these parameters will be included at the end of the returned
        component list.
    """

    macros._verify_sites(subunit, sc_site)
    macros._verify_sites(cargo, c_site)

    def pore_bind_rule_name(rule_expression, size):
        # Get ReactionPatterns
        react_p = rule_expression.reactant_pattern
        prod_p = rule_expression.product_pattern
        # Build the label components
        # Pore is always first complex of LHS due to how we build the rules
        subunit = react_p.complex_patterns[0].monomer_patterns[0].monomer
        if len(react_p.complex_patterns) == 2:
            # This is the complexation reaction
            cargo = react_p.complex_patterns[1].monomer_patterns[0]
        else:
            # This is the dissociation reaction
            cargo = prod_p.complex_patterns[1].monomer_patterns[0]
        return '%s_%d_%s' % (subunit.name, size,
                             macros._monomer_pattern_label(cargo))

    components = ComponentSet()
    # Set up some aliases that are invariant with pore size
    subunit_free = subunit({sc_site: None})
    cargo_free = cargo({c_site: None})

    #for size, klist in zip(range(min_size, max_size + 1), ktable):

    # More aliases which do depend on pore size
    pore_free = macros.pore_species(subunit_free, sp_site1, sp_site2, size)

    # This one is a bit tricky. The pore:cargo complex must only introduce
    # one additional bond even though there are multiple subunits in the
    # pore. We create partial patterns for bound pore and cargo, using a
    # bond number that is high enough not to conflict with the bonds within
    # the pore ring itself.
    # Start by copying pore_free, which has all cargo binding sites empty
    pore_bound = pore_free.copy()
    # Get the next bond number not yet used in the pore structure itself
    cargo_bond_num = size + 1
    # Assign that bond to the first subunit in the pore
    pore_bound.monomer_patterns[0].site_conditions[sc_site] = cargo_bond_num
    # Create a cargo source pattern with that same bond
    cargo_bound = cargo({c_site: cargo_bond_num})
    # Finally we can define the complex trivially; the bond numbers are
    # already present in the patterns
    pc_complex = pore_bound % cargo_bound

    # Create the rules
    name_func = functools.partial(pore_bind_rule_name, size=size)
    components |= macros._macro_rule('pore_bind',
                              pore_free + cargo_free | pc_complex,
                              klist[0:2], ['kf', 'kr'],
                              name_func=name_func)

    return components



from collections import OrderedDict
import networkx as nx
import sympy
import numpy as np
from pysb.simulator import SimulationResult
import matplotlib.colors as colors
import matplotlib.cm as cm
from pyvipr.static_viz import StaticViz, graph_to_json, dot_layout


class OrderedGraph(nx.DiGraph):
    """
    Networkx Digraph that stores the nodes in the order they are input
    """
    node_dict_factory = OrderedDict
    adjlist_outer_dict_factory = OrderedDict
    adjlist_inner_dict_factory = OrderedDict
    edge_attr_dict_factory = OrderedDict


class MidpointNormalize(colors.Normalize):
    """
    A class which, when called, can normalize data into the ``[vmin,midpoint,vmax] interval
    """

    def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
        self.midpoint = midpoint
        colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        # I'm ignoring masked values and all kinds of edge cases to make a
        # simple example...
        x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y))


class ModelVisualization(object):
    """
    Class to visualize the dynamics of systems biology models defined in BNGL, SBML and PySB format.

    Parameters
    ----------
    simulation : pysb SimulationResult
        A SimulationResult instance of the model that is going to be visualized.
    sim_idx : Index of simulation to be visualized
    """
    mach_eps = np.finfo(float).eps

    def __init__(self, simulation, sim_idx=0):
        if not isinstance(simulation, SimulationResult):
            raise TypeError('Argument must be a pysb SimulationResult object')
        self.model = simulation._model
        self.nsims = simulation.nsims
        if self.nsims == 1:
            self.y = simulation.all
        else:
            self.y = simulation.all[sim_idx]
        self.tspan = simulation.tout[sim_idx]
        param_values = simulation.param_values[sim_idx]
        self.param_dict = dict((p.name, param_values[i]) for i, p in enumerate(self.model.parameters))
        self.sp_graph = None
        self.passengers = []
        self.type_viz = ''

    def dynamic_view(self, type_viz='consumption'):
        """
        Generates a dictionary with the model dynamics data that can be converted in the Cytoscape.js JSON format

        Parameters
        ----------
        type_viz : str
            Type of the dynamic visualization, it can be 'consumption' or 'production'

        Examples
        --------
        >>> from pysb.examples.earm_1_0 import model
        >>> from pysb.simulator import ScipyOdeSimulator
        >>> import numpy as np
        >>> tspan = np.linspace(0, 20000)
        >>> sim = ScipyOdeSimulator(model, tspan).run()
        >>> viz = ModelVisualization(sim)
        >>> data = viz.dynamic_view()

        Returns
        -------
        A Dictionary Object with all nodes and edges information that
        can be converted into Cytoscape.js JSON to be visualized
        """
        self.type_viz = type_viz
        self.sp_graph = StaticViz(self.model).species_graph()
        self.sp_graph.graph['view'] = 'dynamic'
        self.sp_graph.graph['nsims'] = self.nsims
        self.sp_graph.graph['tspan'] = self.tspan.tolist()
        g_layout = dot_layout(self.sp_graph)
        self._add_edge_node_dynamics()
        data = graph_to_json(sp_graph=self.sp_graph, layout=g_layout)
        return data

    def dynamic_compartments_view(self, type_viz='consumption'):
        self.type_viz = type_viz
        self.sp_graph = StaticViz(self.model).compartments_data_graph()
        self.sp_graph.graph['view'] = 'dynamic'
        self.sp_graph.graph['nsims'] = self.nsims
        self.sp_graph.graph['tspan'] = self.tspan.tolist()
        g_layout = dot_layout(self.sp_graph)
        self._add_edge_node_dynamics()
        data = graph_to_json(sp_graph=self.sp_graph, layout=g_layout)
        return data

    def dynamic_communities_view(self, type_viz='consumption', random_state=None):
        """

        Parameters
        ----------
        type_viz: str
            Type of visualization. It can be `consumption` to see how species are being consumed
            or `production` to see how the species are being produced.
        random_state: int
            Seed used by the random generator in community detection

        Returns
        -------

        """
        self.type_viz = type_viz
        self.sp_graph = StaticViz(self.model).communities_data_graph(random_state=random_state)
        self.sp_graph.graph['view'] = 'dynamic'
        self.sp_graph.graph['nsims'] = self.nsims
        self.sp_graph.graph['tspan'] = self.tspan.tolist()
        g_layout = dot_layout(self.sp_graph)
        self._add_edge_node_dynamics()
        data = graph_to_json(sp_graph=self.sp_graph, layout=g_layout)
        return data

    def _add_edge_node_dynamics(self):
        """
        Add the edge size and color data as well as node color and values data
        Returns
        -------

        """
        # in networkx 2.0 the order is G, values, names
        # in networkx 1.11 the order is G, names, values
        if float(nx.__version__) >= 2:
            edge_sizes, edge_colors, edge_qtips = self.edges_colors_sizes()
            nx.set_edge_attributes(self.sp_graph, edge_colors, 'edge_color')
            nx.set_edge_attributes(self.sp_graph, edge_sizes, 'edge_size')
            nx.set_edge_attributes(self.sp_graph, edge_qtips, 'edge_qtip')

            node_abs, node_rel = self.node_data()
            nx.set_node_attributes(self.sp_graph, node_abs, 'abs_value')
            nx.set_node_attributes(self.sp_graph, node_rel, 'rel_value')
        else:
            edge_sizes, edge_colors, edge_qtips = self.edges_colors_sizes()
            nx.set_edge_attributes(self.sp_graph, 'edge_color', edge_colors)
            nx.set_edge_attributes(self.sp_graph, 'edge_size', edge_sizes)
            nx.set_edge_attributes(self.sp_graph, 'edge_qtip', edge_qtips)

            node_abs, node_rel = self.node_data()
            nx.set_node_attributes(self.sp_graph, 'abs_value', node_abs)
            nx.set_node_attributes(self.sp_graph, 'rel_value', node_rel)

    def matrix_bidirectional_rates(self):
        """
        Obtains the values of the reaction rates at all the time points of the simulation
        Returns
        -------
        An np.ndarray with the reaction rates values
        """
        rxns_matrix = np.zeros((len(self.model.reactions_bidirectional), len(self.tspan)))

        # Calculates matrix of bidirectional reaction rates
        for idx, reac in enumerate(self.model.reactions_bidirectional):
            rate_reac = reac['rate']
            for p in self.param_dict:
                rate_reac = rate_reac.subs(p, self.param_dict[p])
            variables = [atom for atom in rate_reac.atoms(sympy.Symbol)]
            args = [0] * len(variables)  # arguments to put in the lambdify function
            for idx2, va in enumerate(variables):
                if str(va).startswith('__'):
                    args[idx2] = self.y[str(va)]
                else:
                    args[idx2] = self.param_dict[va.name]
            func = sympy.lambdify(variables, rate_reac, modules=dict(sqrt=np.lib.scimath.sqrt))
            react_rate = func(*args)
            rxns_matrix[idx] = react_rate
        return rxns_matrix

    def edges_colors_sizes(self):
        """
        This function obtains values for the size and color of the edges in the network.
        The color is a representation of the percentage of flux going through an edge.
        The edge size is a representation of the relative value of the reaction normalized to the maximum value that
        the edge can attain during the whole simulation.
        Returns
        -------
        Three dictionaries. The first one contains the information of the edge sizes at all time points.
        The second one contains the information of the edge colors at all time points.
        The third one contains the values of the reaction rates at all time points.
        """
        all_rate_colors = {}
        all_rate_sizes = {}
        all_rate_abs_val = {}

        rxns_matrix = self.matrix_bidirectional_rates()
        vals_norm = np.vectorize(self.mon_normalized)
        all_products = [rx['products'] for rx in self.model.reactions_bidirectional]
        all_reactants = [rx['reactants'] for rx in self.model.reactions_bidirectional]
        sp_imp = set(range(len(self.model.species))) - set(self.passengers)  # species with more than one edge
        # TODO: Make sure that product nodes that also have more than one incoming node, don't overwrite
        #  previous nodes edges attrs

        for sp in sp_imp:
            # Getting the indices of the reactions_bidirectional in which sp is a reactant and a product, respectively
            rxns_idx_reactant = [i for i, rx in enumerate(all_reactants) if sp in rx]
            rxns_idx_product = [i for i, rx in enumerate(all_products) if sp in rx]
            repeated_rxn = set(rxns_idx_reactant) & set(rxns_idx_product)
            # Obtain indices of reactions in which sp is only a product
            rxns_idx_product = [c for c in rxns_idx_product if c not in repeated_rxn]

            # Getting arrays to obtain producing and consuming reactions
            rxn_val_pos = np.where(
                np.concatenate((rxns_matrix[rxns_idx_reactant] > 0, rxns_matrix[rxns_idx_product] < 0)),
                rxns_matrix[rxns_idx_reactant + rxns_idx_product], 0)  # Reactants are consumed
            rxn_val_neg = np.where(
                np.concatenate((rxns_matrix[rxns_idx_reactant] < 0, rxns_matrix[rxns_idx_product] > 0)),
                rxns_matrix[rxns_idx_reactant + rxns_idx_product], 0)  # Reactants are produced

            rxn_val_pos = np.abs(rxn_val_pos)
            rxn_val_neg = np.abs(rxn_val_neg)
            # An array with the total of reaction rates at each time point
            rxn_val_pos_total = rxn_val_pos.sum(axis=0)
            rxn_val_neg_total = rxn_val_neg.sum(axis=0)

            # The max and min of the group of reaction rates. It can be used to assign size to edges
            # based in the max and min of the groups of reaction rates across all time points
            # if rxn_val_pos.size != 0:
            #     rxn_val_pos_max = np.amax(rxn_val_pos) + self.mach_eps
            #
            # if rxn_val_neg.size != 0:
            #     rxn_val_neg_max = np.amax(rxn_val_neg) + self.mach_eps

            sp_rr_dom = rxns_matrix[rxns_idx_reactant]
            sp_prr_dom = rxns_matrix[rxns_idx_product]

            if self.type_viz == 'consumption':
                for rx_idx, rx in enumerate(rxns_idx_reactant):
                    products = all_products[rx]
                    for p in products:
                        # Normalizing by the total flux in a node
                        react_rate_color = rxns_matrix[rx] / rxn_val_pos_total
                        rxn_neg_idx = np.where(react_rate_color < 0)
                        react_rate_color[rxn_neg_idx] = (
                                    rxns_matrix[rx][rxn_neg_idx] / rxn_val_neg_total[rxn_neg_idx])
                        np.nan_to_num(react_rate_color, copy=False)
                        rate_colors = self.f2hex_edges(react_rate_color)

                        rxn_eps = sp_rr_dom[rx_idx] + self.mach_eps  # rxns_matrix[rx] + self.mach_eps
                        react_rate_size = np.zeros(len(rxn_eps))
                        tro_pos_idx = np.where(rxn_eps > 0)[0]
                        tro_neg_idx = np.where(rxn_eps < 0)[0]

                        if tro_neg_idx.size != 0:
                            react_rate_size[tro_neg_idx] = vals_norm(np.abs(rxn_eps[tro_neg_idx]),
                                                                     np.amax(np.abs(rxn_eps[tro_neg_idx])))

                        if tro_pos_idx.size != 0:
                            react_rate_size[tro_pos_idx] = vals_norm(rxn_eps[tro_pos_idx],
                                                                     np.amax(rxn_eps[tro_pos_idx]))

                        rate_sizes = self.range_normalization(react_rate_size, min_x=0, max_x=1)
                        edges_id = ('s{0}'.format(sp), 's{0}'.format(p))
                        all_rate_colors[edges_id] = rate_colors
                        all_rate_sizes[edges_id] = rate_sizes.tolist()
                        all_rate_abs_val[edges_id] = rxns_matrix[rx].tolist()

            elif self.type_viz == 'production':
                for rp_idx, rp in enumerate(rxns_idx_product):
                    reactants = all_reactants[rp]
                    for r in reactants:
                        preact_rate_color = rxns_matrix[rp] / rxn_val_neg_total
                        prxns_neg_idx = np.where(preact_rate_color < 0)
                        preact_rate_color[prxns_neg_idx] = (
                                rxns_matrix[rp][prxns_neg_idx] / rxn_val_pos_total[prxns_neg_idx])
                        prate_colors = self.f2hex_edges(preact_rate_color)

                        prxn_eps = sp_prr_dom[rp_idx] + self.mach_eps
                        preact_rate_size = np.zeros(len(prxn_eps))
                        ptro_pos_idx = np.where(prxn_eps > 0)[0]
                        ptro_neg_idx = np.where(prxn_eps < 0)[0]

                        if ptro_neg_idx.size != 0:
                            preact_rate_size[ptro_neg_idx] = vals_norm(np.abs(prxn_eps[ptro_neg_idx]),
                                                                       np.amax(np.abs(prxn_eps[ptro_neg_idx])))

                        if ptro_pos_idx.size != 0:
                            preact_rate_size[ptro_pos_idx] = vals_norm(prxn_eps[ptro_pos_idx],
                                                                       np.amax(prxn_eps[ptro_pos_idx]))

                        prate_sizes = self.range_normalization(preact_rate_size, min_x=0, max_x=1)
                        edges_id = ('s{0}'.format(r), 's{0}'.format(sp))
                        all_rate_colors[edges_id] = prate_colors
                        all_rate_sizes[edges_id] = prate_sizes.tolist()
                        all_rate_abs_val[edges_id] = rxns_matrix[rp].tolist()
            else:
                raise ValueError('The type of process can only be `consumption` or `production`')

        for sp in self.passengers:
            rxns_idx_reactant = [all_reactants.index(rx) for rx in all_reactants if sp in rx]
            for rx in rxns_idx_reactant:
                products = all_products[rx]
                for p in products:
                    edges_id = ('s{0}'.format(sp), 's{0}'.format(p))
                    all_rate_colors[edges_id] = ['#A4A09F'] * len(self.tspan)
                    all_rate_sizes[edges_id] = [0.5] * len(self.tspan)
                    all_rate_abs_val[edges_id] = rxns_matrix[rx].tolist()

        return all_rate_sizes, all_rate_colors, all_rate_abs_val

    def node_data(self):
        """
        Obtains the species concentration values and the relative concentration
        compared with the maximum concentration across all time points
        Returns
        -------
        Two dictionaries. The first one has the species concentration. The
        second one has the relative species concentrations
        """
        node_absolute = {}
        node_relative = {}
        for sp in range(len(self.model.species)):
            sp_absolute = np.absolute(self.y['__s{}'.format(sp)])
            sp_relative = (sp_absolute / sp_absolute.max()) * 100
            node_absolute['s{0}'.format(sp)] = sp_absolute.tolist()
            node_relative['s{0}'.format(sp)] = sp_relative.tolist()

        # all_nodes_values = pandas.DataFrame(all_rate_colors)
        return node_absolute, node_relative

    @staticmethod
    def f2hex_edges(fx, vmin=-0.99, vmax=0.99):
        """
        Converts reaction rates values to f2hex colors
        Parameters
        ----------
        fx: vector-like
            Vector of reaction rates(flux)
        vmin: float
            Value of the minimum for normalization
        vmax: float
            Value of the maximum for normalization

        Returns
        -------
        A vector of colors in hex format that encodes the reaction rate values
        """
        norm = MidpointNormalize(vmin=vmin, vmax=vmax, midpoint=0)
        f2rgb = cm.ScalarMappable(norm=norm, cmap=cm.get_cmap('RdBu_r'))
        rgb = [f2rgb.to_rgba(rate)[:3] for rate in fx]
        colors_hex = [0] * (len(rgb))
        for i, color in enumerate(rgb):
            colors_hex[i] = '#{0:02x}{1:02x}{2:02x}'.format(*tuple([int(255 * fc) for fc in color]))
        return colors_hex

    @staticmethod
    def range_normalization(x, min_x, max_x, a=0.5, b=10):
        """
        Normalize vector to the [0.5, 10] range
        Parameters
        ----------
        x: vector-like
            Vector of values to be normalized
        min_x: float
            Minimum value in vector x
        max_x: float
            Maximum value in vector x
        a: float
            Value of minimum used for the normalization
        b: float
            Value of maximum used for the normalization

        Returns
        -------
        Normalized vector
        """
        return a + (x - min_x) * (b - a) / (max_x - min_x)

    @staticmethod
    def mon_normalized(x, max_value):
        return x / max_value




"""
Model M1a: Extrinsic apoptosis model with expanded, "embedded together"
model of MOMP.
"""

from pysb import *
from pyvipr.pysb_models import shared
from pyvipr.pysb_models import lopez_modules
from pyvipr.pysb_models import albeck_modules

Model()

# Declare monomers
albeck_modules.ligand_to_c8_monomers()
lopez_modules.momp_monomers()
albeck_modules.apaf1_to_parp_monomers()

# Generate the upstream and downstream sections
albeck_modules.rec_to_bid()
albeck_modules.pore_to_parp()

# The specific MOMP model to use
lopez_modules.embedded()

# Declare observables
shared.observables()




from pysb import *

Model()

Parameter('NA', 6.02e23)
Parameter('kp_LR', 4e8)
Parameter('km_LR', 10)
Parameter('k_LRtrans', 1000)
Parameter('kp_LT1', 4e8)
Parameter('km_LT1', 10)
Parameter('k_LT1trans', 1000)
Parameter('kp_L2T2', 4e8)
Parameter('km_L2T2', 10)
Parameter('k_L2T2trans', 1000)
Parameter('vol_CYT', 1)
Parameter('vol_O1M', 1)
Parameter('vol_O1V', 1)
Parameter('vol_O2M', 1)
Parameter('vol_O2V', 1)
Parameter('L_0', 1000)
Parameter('R_0', 200)
Parameter('T1_0', 1000)
Parameter('T2_0', 1000)

Compartment('CYT', parent=None, dimension=3, size=vol_CYT)
Compartment('OM1', parent=CYT, dimension=2, size=vol_O1M)
Compartment('OV1', parent=OM1, dimension=3, size=vol_O1V)
Compartment('OM2', parent=CYT, dimension=2, size=vol_O2M)
Compartment('OV2', parent=OM2, dimension=3, size=vol_O2V)

Monomer('L')
Monomer('R')
Monomer('LR')
Monomer('T1')
Monomer('LT1')
Monomer('L2')
Monomer('T2')
Monomer('L2T2')

Initial(L() ** CYT, L_0)
Initial(R() ** OM1, R_0)
Initial(T1() ** OM1, T1_0)
Initial(T2() ** OM2, T2_0)

Rule('bind_L_R', L() ** CYT + R() ** OM1 | LR() ** OM1, kp_LR, km_LR)
Rule('L_transport_OV1', LR() ** OM1 >> L() ** OV1 + R() ** OM1, k_LRtrans)
Rule('bind_L_T1', L() ** OV1 + T1() ** OM1 | LT1() ** OM1, kp_LT1, km_LT1)
Rule('create_L2', LT1() ** OM1 >> T1() ** OM1 + L2() ** CYT, k_LT1trans)
Rule('bind_L2_T2', L2() ** CYT + T2() ** OM2 | L2T2() ** OM2, kp_L2T2, km_L2T2)
Rule('L2_transport_OV2', L2T2() ** OM2 >> T2() ** OM2 + L2() ** OV2, k_L2T2trans)

# L@CYT + R@O1M <-> LR@O1M kp_LR, km_LR
# LR@O1M -> L@O1V + R@O1M k_LRtrans
# L@O1V + T1@O1M <-> LT1@O1M kp_LT1, km_LT1
# LT1@O1M -> T1@O1M + L2@CYT k_LT1trans
# L2@CYT + T2@O2M <-> L2T2@O2M kp_L2T2, km_L2T2
# L2T2@O2M -> T2@O2M + L2@O2V k_L2T2trans